"""
Contains code for calculating and writing image header information.
"""

import os
import math

import azcam


class ImageHeaders(object):
    """
    Class for image headers.
    """

    def __init__(self):
        super().__init__()

    def _write_PHU(self, hdu):
        """
        Write primary header for FITS or MEF file.
        """

        numHDUs = self.focalplane.numamps_image
        if numHDUs == 1:  # no extensions for single amp
            numHDUs = 0
        hdu.header.set(
            "NEXTEND", numHDUs, "Number of extensions"
        )  # near top of FITS header
        hdu.header.set(
            "BITPIX", 16, "array data type"
        )  # so 8 bits never shows up in PHU

        # DETSIZE is whole mosaic size or CCD if single device (unbinned)
        x = (
            self.focalplane.ampvispix_x
            * self.focalplane.num_ser_amps_det
            * self.focalplane.numdet_x
        )
        y = (
            self.focalplane.ampvispix_y
            * self.focalplane.num_par_amps_det
            * self.focalplane.numdet_y
        )
        s = "[1:%d,1:%d]" % (x, y)
        hdu.header.set("DETSIZE", s, "Detector size")

        # CCDSUM is binning
        s = "%d %d" % (self.focalplane.col_bin, self.focalplane.row_bin)
        hdu.header.set("CCDSUM", s, "CCD pixel summing")
        hdu.header.set(
            "CCDBIN1", self.focalplane.col_bin, "Binning factor along axis 1"
        )
        hdu.header.set(
            "CCDBIN2", self.focalplane.row_bin, "Binning factor along axis 2"
        )

        # filename at acquisition (no folder)
        filename = self.filename
        filename = os.path.basename(filename)
        hdu.header.set("FILENAME", filename, "base filename at acquisition")

        hdu.header.set("NCCDS", self.focalplane.num_detectors, "Number of CCDs")
        hdu.header.set("NAMPS", self.focalplane.numamps_image, "Number of amplifiers")

        # keywords are written after exposure is completed
        curpos = len(hdu.header)
        for headername in azcam.db.headerorder:
            item_header = azcam.db.headers[headername]
            cheader = item_header.get_header()  # list of kw, value, comment, type]
            if cheader == [] or cheader is None:
                continue
            # first add the comment lines
            for comm in item_header.title:
                hdu.header.add_comment(item_header.title[comm], after=curpos)
                curpos = curpos + 1

            # add the keywords
            for head in cheader:
                if head[0].lower() == "comment":
                    hdu.header.add_comment(head[1], after=curpos)
                elif head[0].lower() == "history":
                    hdu.header.add_history(head[1], after=curpos)
                else:
                    try:
                        hdu.header.set(head[0], head[1], head[2], after=curpos)
                    except Exception:
                        curpos -= 1
                curpos += 1

        return

    def _write_asm_fits_header(self, hdu):
        """
        Write primary header for assembled FITS file.
        """

        # no extensions
        numHDUs = 2

        hdu.header.set("EXTEND", True, "")
        hdu.header.set("NEXTEND", numHDUs, "Number of extensions")
        hdu.header.set("BITPIX", 16, "array data type")

        # DETSIZE is whole mosaic size or CCD if single device (unbinned)
        x = self.asmsize[0]
        y = self.asmsize[1]
        s = "[1:%d,1:%d]" % (x, y)
        hdu.header.set("DETSIZE", s, "Detector size")

        # CCDSUM is binning
        s = "%d %d" % (self.focalplane.col_bin, self.focalplane.row_bin)
        hdu.header.set("CCDSUM", s, "CCD pixel summing")
        hdu.header.set(
            "CCDBIN1", self.focalplane.col_bin, "Binning factor along axis 1"
        )
        hdu.header.set(
            "CCDBIN2", self.focalplane.row_bin, "Binning factor along axis 2"
        )

        # filename at acquisition (no folder)
        filename = self.filename
        filename = os.path.basename(filename)
        hdu.header.set("FILENAME", filename, "base filename at acquisition")

        hdu.header.set("NCCDS", 1, "Number of CCDs")
        hdu.header.set("NAMPS", 1, "Number of amplifiers")

        # all these keywords are written after exposure is finished
        curpos = len(hdu.header)
        for item in self.header.items:
            cheader = item.header.get_header()
            if cheader == []:
                continue
            # first add the comment lines
            for comm in item.header.title:
                # add the comment line
                hdu.header.add_comment(item.header.title[comm], after=curpos)
                curpos = curpos + 1

            # add the keywords
            for head in cheader:
                if head[0].lower() == "comment":
                    hdu.header.add_comment(head[1], after=curpos)
                elif head[0].lower() == "history":
                    hdu.header.add_history(head[1], after=curpos)
                else:
                    hdu.header.set(head[0], head[1], head[2], after=curpos)
                curpos += 1

        return

    def _write_extension_header(self, ext_number, hdu):
        """
        Write the extension header for a FITS/MEF file.
        ext_number starts at 1 (for MEF and FITS).
        """

        # make nice header text
        curpos = len(hdu.header)
        hdu.header.add_comment(
            "==================================================================",
            after=curpos,
        )
        hdu.header.add_comment("Image", after=curpos + 1)
        hdu.header.add_comment(
            "==================================================================",
            after=curpos + 2,
        )
        curpos = curpos + 3

        # DATASEC is actual DATA BINNED pixels in the final image
        # does not include over/under scans
        self.datasecx1 = self.focalplane.xunderscan + 1
        self.datasecx2 = self.focalplane.xunderscan + self.focalplane.numviscols_amp
        self.datasecy1 = self.focalplane.yunderscan + 1
        self.datasecy2 = self.focalplane.yunderscan + self.focalplane.numvisrows_amp

        # BIASSEC is over or under scan - based on its location
        # defaults to DATASEC if none requested
        self.biassecx1 = self.datasecx1
        self.biassecx2 = self.datasecx2
        self.biassecy1 = self.datasecy1
        self.biassecy2 = self.datasecy2

        if self.focalplane.bias_location == self.focalplane.BL_COLUNDER:
            self.biassecx1 -= self.focalplane.xunderscan
            self.biassecx2 -= self.focalplane.numviscols_amp
        elif self.focalplane.bias_location == self.focalplane.BL_COLOVER:
            self.biassecx1 += self.focalplane.numviscols_amp
            self.biassecx2 += self.focalplane.xoverscan
        elif self.focalplane.bias_location == self.focalplane.BL_ROWUNDER:
            self.biassecy1 -= self.focalplane.yunderscan
            self.biassecy2 -= self.focalplane.numvisrows_amp
        elif self.focalplane.bias_location == self.focalplane.BL_ROWOVER:
            self.biassecy1 += self.focalplane.numvisrows_amp
            self.biassecy2 += self.focalplane.yoverscan

        # TRIMSEC is DATASEC, the pixels to keep after removing BIASSEC
        self.trimsecx1 = self.datasecx1
        self.trimsecx2 = self.datasecx2
        self.trimsecy1 = self.datasecy1
        self.trimsecy2 = self.datasecy2

        # CCDSEC is unbinned
        # new code - unbinned
        self.ccdsecX1 = self.datasecx1
        self.ccdsecX2 = self.focalplane.ampvispix_x
        self.ccdsecY1 = self.datasecy1
        self.ccdsecY2 = self.focalplane.ampvispix_y

        # old code - binned
        self.ccdsecx1 = self.datasecx1
        self.ccdsecx2 = self.datasecx2
        self.ccdsecy1 = self.datasecy1
        self.ccdsecy2 = self.datasecy2

        # ORIGSEC is entire device in binned coordinates
        self.origsecx1 = 1
        # self.origsecx2	= self.coltotal/self.col_bin GSZ
        self.origsecx2 = (
            self.focalplane.numcols_amp
            * self.focalplane.num_ser_amps_det
            / self.focalplane.col_bin
        )
        self.origsecy1 = 1
        # self.origsecy2	= self.rowtotal/self.row_bin
        self.origsecy2 = (
            self.focalplane.numrows_amp
            * self.focalplane.num_par_amps_det
            / self.focalplane.row_bin
        )

        # determine amplifier orientation: new version GSZ 04.14.2011
        ampflip = self.focalplane.amp_cfg[ext_number - 1]

        # calculate which part of mosaic from HDU (increments first in X)
        ny = self.focalplane.extpos_y[ext_number - 1]
        nx = self.focalplane.extpos_x[ext_number - 1]
        det_number = self.focalplane.det_number[ext_number - 1]
        hdu.header.set(
            "IMAGEID",
            self.focalplane.ext_number[ext_number - 1],
            "Image ID",
            after=curpos,
        )  # int

        curpos += 1

        s = "ccd%d" % det_number
        hdu.header.set("CCDNAME", s, "CCD name", after=curpos)  # string
        curpos += 1

        # all unbinned
        # DETECTOR refers to the mosaic
        # CCD refers to the individual detector
        # AMP refers to data from one amplifier
        # DETSIZE is entire image size

        xVal = (
            self.focalplane.ampvispix_x
            * self.focalplane.num_ser_amps_det
            * self.focalplane.numdet_x
        )
        yVal = (
            self.focalplane.ampvispix_y
            * self.focalplane.num_par_amps_det
            * self.focalplane.numdet_y
        )
        s = "[1:%d,1:%d]" % (xVal, yVal)
        hdu.header.set("DETSIZE", s, "Detector size", after=curpos)  # string
        curpos += 1

        # CCDSIZE is the image size on each CCD, unbinned
        xVal = self.focalplane.ampvispix_x * self.focalplane.num_ser_amps_det
        yVal = self.focalplane.ampvispix_y * self.focalplane.num_par_amps_det
        s = "[1:%d,1:%d]" % (xVal, yVal)
        hdu.header.set("CCDSIZE", s, "CCD size", after=curpos)  # string
        curpos += 1

        # write the extension coordinate keywords
        if self.focalplane.numcols_overscan > 0:
            s = "[%d:%d,%d:%d]" % (
                self.focalplane.numviscols_amp + 1,
                self.focalplane.numviscols_amp + self.focalplane.numcols_overscan,
                1,
                self.focalplane.numvisrows_amp,
            )
        else:
            s = "[1:%d,1:%d]" % (
                self.focalplane.numviscols_amp,
                self.focalplane.numvisrows_amp,
            )
        hdu.header.set("BIASSEC", s, "Bias section", after=curpos)  # string
        curpos += 1

        # DATASEC is data region
        s = "[%d:%d,%d:%d]" % (
            self.datasecx1,
            self.datasecx2,
            self.datasecy1,
            self.datasecy2,
        )
        hdu.header.set("DATASEC", s, "Data section", after=curpos)  # string
        curpos += 1

        # TRIMSEC is trim region for trimming and display
        s = "[%d:%d,%d:%d]" % (
            self.trimsecx1,
            self.trimsecx2,
            self.trimsecy1,
            self.trimsecy2,
        )
        hdu.header.set("TRIMSEC", s, "Trim section", after=curpos)  # string
        curpos += 1

        # AMPSEC is the data from one amp (DATASEC for one amp), unbinned
        if ampflip == 0:
            # no flip
            s = "[%d:%d,%d:%d]" % (
                self.ccdsecX1,
                self.ccdsecX2,
                self.ccdsecY1,
                self.ccdsecY2,
            )
        elif ampflip == 1:
            # flip in x
            s = "[%d:%d,%d:%d]" % (
                self.ccdsecX2,
                self.ccdsecX1,
                self.ccdsecY1,
                self.ccdsecY2,
            )
        elif ampflip == 2:
            # flip in y
            s = "[%d:%d,%d:%d]" % (
                self.ccdsecX1,
                self.ccdsecX2,
                self.ccdsecY2,
                self.ccdsecY1,
            )
        elif ampflip == 3:
            # flip both
            s = "[%d:%d,%d:%d]" % (
                self.ccdsecX2,
                self.ccdsecX1,
                self.ccdsecY2,
                self.ccdsecY1,
            )

        hdu.header.set("AMPSEC", s, "Amplifier section", after=curpos)  # string
        curpos += 1

        # CCDSEC is not the same as DETSEC for a mosaic

        # amplifier's positions for CCDSEC
        if self.focalplane.detpos_x[ext_number - 1] > 1:
            Nx = (
                self.focalplane.extpos_x[ext_number - 1]
                - self.focalplane.detpos_x[ext_number - 1]
            )
        else:
            Nx = nx

        if self.focalplane.detpos_y[ext_number - 1] > 1:
            Ny = (
                self.focalplane.extpos_y[ext_number - 1]
                - self.focalplane.detpos_y[ext_number - 1]
            )
        else:
            Ny = ny

        # DETSEC is the region from the mosaic where this data is from
        # defines how image is displayed and includes flips

        if self.focalplane.numamps_x == 1:
            lastCol = self.focalplane.last_col
        else:
            lastCol = self.focalplane.ampvispix_x

        if self.focalplane.numamps_y == 1:
            lastRow = self.focalplane.last_row
        else:
            lastRow = self.focalplane.ampvispix_y

        skipX1 = self.focalplane.first_col - 1
        skipX2 = self.focalplane.ampvispix_x - self.focalplane.last_col
        skipX2 = max(0, skipX2)
        skipY1 = self.focalplane.first_row - 1
        skipY2 = self.focalplane.ampvispix_y - self.focalplane.last_row
        skipY2 = max(0, skipY2)

        skipX1bin = skipX1 / self.focalplane.col_bin
        skipY1bin = skipY1 / self.focalplane.row_bin

        if ampflip == 0:  # no flip
            flip_x = 1
            flip_y = 1

            # DETSEC
            xVal1 = (nx - 1) * self.focalplane.ampvispix_x + self.focalplane.first_col
            xVal2 = nx * lastCol
            yVal1 = (ny - 1) * self.focalplane.ampvispix_y + self.focalplane.first_row
            yVal2 = ny * lastRow

            # CCDSEC
            X_Val1 = (Nx - 1) * self.focalplane.ampvispix_x + self.focalplane.first_col
            X_Val2 = Nx * lastCol
            Y_Val1 = (Ny - 1) * self.focalplane.ampvispix_y + self.focalplane.first_row
            Y_Val2 = Ny * lastRow

            # CCDSEC1 binned version of CCDSEC
            xCCD1 = skipX1bin + 1
            xCCD2 = (self.focalplane.ampvispix_x - skipX1) / self.focalplane.col_bin
            yCCD1 = skipY1bin + 1
            yCCD2 = (self.focalplane.ampvispix_y - skipY2) / self.focalplane.row_bin

        elif ampflip == 1:  # flip in x
            flip_x = -1
            flip_y = 1

            # DETSEC
            xVal1 = nx * lastCol
            xVal2 = (nx - 1) * self.focalplane.ampvispix_x + self.focalplane.first_col
            yVal1 = (ny - 1) * self.focalplane.ampvispix_y + self.focalplane.first_row
            yVal2 = ny * lastRow

            # CCDSEC
            X_Val1 = Nx * lastCol
            X_Val2 = (Nx - 1) * self.focalplane.ampvispix_x + self.focalplane.first_col
            Y_Val1 = (Ny - 1) * self.focalplane.ampvispix_y + self.focalplane.first_row
            Y_Val2 = Ny * lastRow

            # CCDSEC1 binned version of CCDSEC
            xCCD1 = (
                (Nx * self.focalplane.ampvispix_x) - skipX1
            ) / self.focalplane.col_bin
            # xCCD2 = (
            #     (Nx - 1) * self.focalplane.ampvispix_x - skipX1
            # ) / self.focalplane.col_bin + 1
            xCCD2 = (
                (Nx - 1) * self.focalplane.ampvispix_x - skipX1
            ) / self.focalplane.col_bin
            yCCD1 = skipY1bin + 1
            yCCD2 = (self.focalplane.ampvispix_y - skipY2) / self.focalplane.row_bin

        elif ampflip == 2:  # flip in y
            flip_x = 1
            flip_y = -1

            # DETSEC
            xVal1 = (nx - 1) * self.focalplane.ampvispix_x + self.focalplane.first_col
            xVal2 = nx * lastCol
            yVal1 = ny * lastRow
            yVal2 = (ny - 1) * self.focalplane.ampvispix_y + self.focalplane.first_row

            # CCDSEC
            X_Val1 = (Nx - 1) * self.focalplane.ampvispix_x + self.focalplane.first_col
            X_Val2 = Nx * lastCol
            Y_Val1 = Ny * lastRow
            Y_Val2 = (Ny - 1) * self.focalplane.ampvispix_y + self.focalplane.first_row

            # CCDSEC1 binned version of CCDSEC
            xCCD1 = skipX1bin + 1
            xCCD2 = (self.focalplane.ampvispix_x - skipX1) / self.focalplane.col_bin
            yCCD1 = (
                (Ny * self.focalplane.ampvispix_y) - skipY1
            ) / self.focalplane.row_bin
            # yCCD2 = (
            #     (Ny - 1) * self.focalplane.ampvispix_y - skipY1
            # ) / self.focalplane.row_bin + 1
            yCCD2 = (
                (Ny - 1) * self.focalplane.ampvispix_y - skipY1
            ) / self.focalplane.row_bin

        elif ampflip == 3:  # flip both
            flip_x = -1
            flip_y = -1

            # DETSEC
            xVal1 = nx * lastCol
            xVal2 = (nx - 1) * self.focalplane.ampvispix_x + self.focalplane.first_col
            yVal1 = ny * lastRow
            yVal2 = (ny - 1) * self.focalplane.ampvispix_y + self.focalplane.first_row

            # CCDSEC
            X_Val1 = Nx * lastCol
            X_Val2 = (Nx - 1) * self.focalplane.ampvispix_x + self.focalplane.first_col
            Y_Val1 = Ny * lastRow
            Y_Val2 = (Ny - 1) * self.focalplane.ampvispix_y + self.focalplane.first_row

            # CCDSEC1 binned version of CCDSEC
            # xCCD1 = (
            #     (Nx - 1) * self.focalplane.ampvispix_x - skipX1
            # ) / self.focalplane.col_bin + 1
            xCCD1 = (
                (Nx - 1) * self.focalplane.ampvispix_x - skipX1
            ) / self.focalplane.col_bin
            xCCD2 = (
                (Nx * self.focalplane.ampvispix_x) - skipX1
            ) / self.focalplane.col_bin
            yCCD1 = (
                (Ny * self.focalplane.ampvispix_y) - skipY1
            ) / self.focalplane.row_bin
            # yCCD2 = (
            #     (Ny - 1) * self.focalplane.ampvispix_y - skipY1
            # ) / self.focalplane.row_bin + 1
            yCCD2 = (
                (Ny - 1) * self.focalplane.ampvispix_y - skipY1
            ) / self.focalplane.row_bin

        # do not overwrite x_Val1 and y_Val2; these are needed to calculate LTV1 and LTV2
        s = "[%d:%d,%d:%d]" % (xVal1, xVal2, yVal1, yVal2)
        hdu.header.set("DETSEC", s, "Detector section", after=curpos)
        curpos += 1

        s = "[%d:%d,%d:%d]" % (X_Val1, X_Val2, Y_Val1, Y_Val2)
        hdu.header.set("CCDSEC", s, "CCD section", after=curpos)
        curpos += 1

        xCCD1 = max(1, xCCD1)
        xCCD2 = max(1, xCCD2)
        yCCD1 = max(1, yCCD1)
        yCCD2 = max(1, yCCD2)
        s = "[%d:%d,%d:%d]" % (xCCD1, xCCD2, yCCD1, yCCD2)
        hdu.header.set("CCDSEC1", s, "CCD section with binning", after=curpos)

        curpos += 1

        # other versions of same info
        hdu.header.set(
            "OVRSCAN1", self.focalplane.xoverscan, "Overscan on axis 1", after=curpos
        )
        curpos += 1
        hdu.header.set(
            "OVRSCAN2", self.focalplane.yoverscan, "Overscan on axis 2", after=curpos
        )
        curpos += 1
        hdu.header.set(
            "PRESCAN1", self.focalplane.xunderscan, "Underscan on axis 1", after=curpos
        )
        curpos += 1
        hdu.header.set(
            "PRESCAN2", self.focalplane.yunderscan, "Underscan on axis 2", after=curpos
        )
        curpos += 1

        # include CCDSUM
        s = "%s %s" % (self.focalplane.col_bin, self.focalplane.row_bin)
        hdu.header.set("CCDSUM", s, "CCD pixel summing", after=curpos)
        curpos += 1

        x_Val1 = float(xVal1)
        y_Val1 = float(yVal1)
        y_Val2 = float(yVal2)
        x_Val2 = float(xVal2)  # ? added 07sep24

        if self.focalplane.numamps_image > 1:
            self.focalplane.wcs.ltm_1_1[ext_number - 1] = flip_x / float(
                self.focalplane.col_bin
            )
            self.focalplane.wcs.ltm_2_2[ext_number - 1] = flip_y / float(
                self.focalplane.row_bin
            )

            if self.focalplane.split_physical_coords == 1:
                # split physical coordinates
                self.focalplane.wcs.ltv_1[ext_number - 1] = (
                    -flip_x * X_Val1 + 1
                ) / float(self.focalplane.col_bin)
                self.focalplane.wcs.ltv_2[ext_number - 1] = (
                    -flip_y * Y_Val1 + 1
                ) / float(self.focalplane.row_bin)
            else:
                # combine physical coordinates
                self.focalplane.wcs.ltv_1[ext_number - 1] = (
                    -flip_x * x_Val1 + 1
                ) / float(self.focalplane.col_bin)
                self.focalplane.wcs.ltv_2[ext_number - 1] = (
                    -flip_y * y_Val1 + 1
                ) / float(self.focalplane.row_bin)

            # include LTM1_1, LTM2_2, LTV1, and LTV2
            hdu.header.set(
                "LTM1_1",
                self.focalplane.wcs.ltm_1_1[ext_number - 1],
                "CCD to image transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "LTM2_2",
                self.focalplane.wcs.ltm_2_2[ext_number - 1],
                "CCD to image transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "LTV1",
                self.focalplane.wcs.ltv_1[ext_number - 1],
                "CCD to image transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "LTV2",
                self.focalplane.wcs.ltv_2[ext_number - 1],
                "CCD to image transformation",
                after=curpos,
            )
            curpos += 1

            self.focalplane.wcs.atm_1_1[ext_number - 1] = flip_x
            self.focalplane.wcs.atm_2_2[ext_number - 1] = flip_y

            Ns = self.focalplane.ampvispix_x
            Np = self.focalplane.ampvispix_y

            if flip_x == -1:
                ext = 1
            else:
                ext = 0

            if self.focalplane.split_physical_coords == 1:
                # ATV1 for split physical coordinates
                if self.focalplane.wcs.atm_1_1[ext_number - 1] == 1:
                    self.focalplane.wcs.atv1[ext_number - 1] = -(Ns * (Nx - 1) + ext)
                else:
                    self.focalplane.wcs.atv1[ext_number - 1] = Ns * Nx + ext
            else:
                # ATV1 for combined physical coordinates
                if self.focalplane.wcs.atm_1_1[ext_number - 1] == 1:
                    self.focalplane.wcs.atv1[ext_number - 1] = -(
                        Ns * (self.focalplane.extpos_x[ext_number - 1] - 1) + ext
                    )
                else:
                    self.focalplane.wcs.atv1[ext_number - 1] = (
                        Ns * self.focalplane.extpos_x[ext_number - 1] + ext
                    )

            if flip_y == -1:
                ext = 1
            else:
                ext = 0

            if self.focalplane.split_physical_coords == 1:
                # ATV2 for split physical coordinates
                if self.focalplane.wcs.atm_2_2[ext_number - 1] == 1:
                    self.focalplane.wcs.atv2[ext_number - 1] = -(Np * (Ny - 1) + ext)
                else:
                    self.focalplane.wcs.atv2[ext_number - 1] = Np * Ny + ext
            else:
                # ATV2 for combined physical coordinates
                if self.focalplane.wcs.atm_2_2[ext_number - 1] == 1:
                    self.focalplane.wcs.atv2[ext_number - 1] = -(
                        Np * (self.focalplane.extpos_y[ext_number - 1] - 1) + ext
                    )
                else:
                    self.focalplane.wcs.atv2[ext_number - 1] = (
                        Np * self.focalplane.extpos_y[ext_number - 1] + ext
                    )

            # include ATM1_1, ATM2_2, ATV1, and ATV2
            hdu.header.set(
                "ATM1_1",
                self.focalplane.wcs.atm_1_1[ext_number - 1],
                "CCD to amplifier transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "ATM2_2",
                self.focalplane.wcs.atm_2_2[ext_number - 1],
                "CCD to amplifier transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "ATV1",
                self.focalplane.wcs.atv1[ext_number - 1],
                "CCD to amplifier transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "ATV2",
                self.focalplane.wcs.atv2[ext_number - 1],
                "CCD to amplifier transformation",
                after=curpos,
            )
            curpos += 1

            # calculate CCD to detector transformation matrix and vectors

            # DTV1 and DTV2 set to 0 combines physical coordinates for all CCDs
            self.focalplane.wcs.dtm_1_1[ext_number - 1] = 1
            self.focalplane.wcs.dtm_2_2[ext_number - 1] = 1

            if self.focalplane.split_physical_coords == 1:
                # DTV1 and DTV2 for split physical coordinates
                self.focalplane.wcs.dtv_1[ext_number - 1] = (
                    (self.focalplane.detpos_x[ext_number - 1] - 1)
                    * self.focalplane.ampvispix_x
                    * self.focalplane.num_par_amps_det
                )
                self.focalplane.wcs.dtv_2[ext_number - 1] = (
                    (self.focalplane.detpos_y[ext_number - 1] - 1)
                    * self.focalplane.ampvispix_y
                    * self.focalplane.num_ser_amps_det
                )
            else:
                # DTV1 and DTV2 for combined physical coordinates
                self.focalplane.wcs.dtv_1[ext_number - 1] = 0
                self.focalplane.wcs.dtv_2[ext_number - 1] = 0

            # include DTM1_1, DTM2_2, DTV1, and DTV2
            hdu.header.set(
                "DTM1_1",
                self.focalplane.wcs.dtm_1_1[ext_number - 1],
                "CCD to detector transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "DTM2_2",
                self.focalplane.wcs.dtm_2_2[ext_number - 1],
                "CCD to detector transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "DTV1",
                self.focalplane.wcs.dtv_1[ext_number - 1],
                "CCD to detector transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "DTV2",
                self.focalplane.wcs.dtv_2[ext_number - 1],
                "CCD to detector transformation",
                after=curpos,
            )
            curpos += 1
        else:
            # detectors with single amplifer

            # calculate CCD to image transformation matrix and vectors
            self.focalplane.wcs.ltm_1_1[ext_number - 1] = flip_x / float(
                self.focalplane.col_bin
            )
            self.focalplane.wcs.ltm_2_2[ext_number - 1] = flip_y / float(
                self.focalplane.row_bin
            )

            self.focalplane.wcs.ltv_1[ext_number - 1] = (
                self.datasecx1
                - self.focalplane.wcs.ltm_1_1[ext_number - 1] * x_Val1
                - 0.5 * (1 - self.focalplane.wcs.ltm_1_1[ext_number - 1])
            )
            self.focalplane.wcs.ltv_2[ext_number - 1] = (
                self.datasecy2
                - self.focalplane.wcs.ltm_2_2[ext_number - 1] * y_Val2
                - 0.5 * (1 - self.focalplane.wcs.ltm_2_2[ext_number - 1])
            )

            # include LTM1_1, LTM2_2, LTV1, and LTV2
            hdu.header.set(
                "LTM1_1",
                self.focalplane.wcs.ltm_1_1[ext_number - 1],
                "CCD to image transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "LTM2_2",
                self.focalplane.wcs.ltm_2_2[ext_number - 1],
                "CCD to image transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "LTV1",
                self.focalplane.wcs.ltv_1[ext_number - 1],
                "CCD to image transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "LTV2",
                self.focalplane.wcs.ltv_2[ext_number - 1],
                "CCD to image transformation",
                after=curpos,
            )
            curpos += 1

            self.focalplane.wcs.atm_1_1[ext_number - 1] = flip_x
            self.focalplane.wcs.atm_2_2[ext_number - 1] = flip_y

            Ns = self.focalplane.ampvispix_x
            Np = self.focalplane.ampvispix_y

            if Nx > 1:
                ext = 1
            else:
                ext = 0

            if self.focalplane.wcs.atm_1_1[ext_number - 1] == 1:
                self.focalplane.wcs.atv1[ext_number - 1] = -(
                    Ns * (self.focalplane.extpos_x[ext_number - 1] - 1) + ext
                )
            else:
                self.focalplane.wcs.atv1[ext_number - 1] = (
                    Ns * self.focalplane.extpos_x[ext_number - 1] + ext
                )

            if Ny > 1:
                ext = 1
            else:
                ext = 0

            if self.focalplane.wcs.atm_2_2[ext_number - 1] == 1:
                self.focalplane.wcs.atv2[ext_number - 1] = -(
                    Np * (self.focalplane.extpos_y[ext_number - 1] - 1) + ext
                )
            else:
                self.focalplane.wcs.atv2[ext_number - 1] = (
                    Np * self.focalplane.extpos_y[ext_number - 1] + ext
                )

            hdu.header.set(
                "ATM1_1",
                self.focalplane.wcs.atm_1_1[ext_number - 1],
                "CCD to amplifier transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "ATM2_2",
                self.focalplane.wcs.atm_2_2[ext_number - 1],
                "CCD to amplifier transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "ATV1",
                self.focalplane.wcs.atv1[ext_number - 1],
                "CCD to amplifier transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "ATV2",
                self.focalplane.wcs.atv2[ext_number - 1],
                "CCD to amplifier transformation",
                after=curpos,
            )
            curpos += 1

            # DTV1 and DTV2 are set to 0 since there is only one CCD
            self.focalplane.wcs.dtm_1_1[ext_number - 1] = 1
            self.focalplane.wcs.dtm_2_2[ext_number - 1] = 1

            self.focalplane.wcs.dtv_1[ext_number - 1] = 0
            self.focalplane.wcs.dtv_2[ext_number - 1] = 0

            hdu.header.set(
                "DTM1_1",
                self.focalplane.wcs.dtm_1_1[ext_number - 1],
                "CCD to detector transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "DTM2_2",
                self.focalplane.wcs.dtm_2_2[ext_number - 1],
                "CCD to detector transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "DTV1",
                self.focalplane.wcs.dtv_1[ext_number - 1],
                "CCD to detector transformation",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "DTV2",
                self.focalplane.wcs.dtv_2[ext_number - 1],
                "CCD to detector transformation",
                after=curpos,
            )
            curpos += 1

        return

    def _write_wcs_keywords(self, ext_number, hdu):
        """
        Define and write the WCS keywords to the header.
        ext_number starts at 1 (for MEF and FITS).
        """

        if not self.write_wcs:
            return

        curpos = len(hdu.header)
        hdu.header.add_comment(
            "==================================================================",
            after=curpos,
        )
        hdu.header.add_comment("WCS", after=curpos + 1)
        hdu.header.add_comment(
            "==================================================================",
            after=curpos + 2,
        )
        curpos = curpos + 3
        hdu.header.set(
            "EQUINOX", self.focalplane.wcs.equinox, "Equinox of WCS", after=curpos
        )
        curpos += 1
        hdu.header.set(
            "WCSDIM", self.focalplane.wcs.wcs_dim, "WCS Dimensionality", after=curpos
        )
        curpos += 1
        hdu.header.set(
            "CTYPE1", self.focalplane.wcs.ctype1, "Coordinate type", after=curpos
        )
        curpos += 1
        hdu.header.set(
            "CTYPE2", self.focalplane.wcs.ctype2, "Coordinate type", after=curpos
        )
        curpos += 1

        self.focalplane.wcs.get_ra_dec()

        if self.focalplane.wcs.ctype1.startswith("RA"):
            hdu.header.set(
                "CRVAL1",
                self.focalplane.wcs.ra_deg,
                "Coordinate reference value",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "CRVAL2",
                self.focalplane.wcs.dec_deg,
                "Coordinate reference value",
                after=curpos,
            )
        else:
            hdu.header.set(
                "CRVAL1",
                self.focalplane.wcs.dec_deg,
                "Coordinate reference value",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "CRVAL2",
                self.focalplane.wcs.ra_deg,
                "Coordinate reference value",
                after=curpos,
            )

        curpos += 1

        ampflip = self.focalplane.amp_cfg[ext_number - 1]  # from to 0 through 3
        if ampflip == 0:  # no flip
            flip_x = 1
            flip_y = 1
        elif ampflip == 1:  # flip in x
            flip_x = -1
            flip_y = 1
        elif ampflip == 2:  # flip in y
            flip_x = 1
            flip_y = -1
        elif ampflip == 3:  # flip both
            flip_x = -1
            flip_y = -1

        ref1 = (
            self.focalplane.refpix1 - self.focalplane.amppix1[ext_number - 1]
        ) * self.focalplane.wcs.atm_1_1[ext_number - 1]
        ref1 = (ref1 - (self.focalplane.first_col - 1)) / self.focalplane.col_bin

        ref2 = (
            self.focalplane.refpix2 - self.focalplane.amppix2[ext_number - 1]
        ) * self.focalplane.wcs.atm_2_2[ext_number - 1]
        ref2 = (ref2 - (self.focalplane.first_row - 1)) / self.focalplane.row_bin

        hdu.header.set("CRPIX1", ref1, "Coordinate reference pixel", after=curpos)
        curpos += 1
        hdu.header.set("CRPIX2", ref2, "Coordinate reference pixel", after=curpos)
        curpos += 1

        if self.focalplane.wcs.cd_matrix == 0:
            alfa = self.focalplane.wcs.rot_deg[ext_number - 1]

            CD1_1 = (
                self.focalplane.wcs.scale1[ext_number - 1]
                * self.focalplane.col_bin
                * math.cos(alfa * 2 * math.pi / 360.0)
            ) * flip_x
            CD1_2 = (
                self.focalplane.wcs.scale2[ext_number - 1]
                * self.focalplane.row_bin
                * (-math.sin(alfa * 2 * math.pi / 360.0))
            ) * flip_y
            CD2_1 = (
                self.focalplane.wcs.scale1[ext_number - 1]
                * self.focalplane.col_bin
                * math.sin(alfa * 2 * math.pi / 360.0)
            ) * flip_x
            CD2_2 = (
                self.focalplane.wcs.scale2[ext_number - 1]
                * self.focalplane.row_bin
                * math.cos(alfa * 2 * math.pi / 360.0)
            ) * flip_y

            hdu.header.set("CD1_1", CD1_1, "Coordinate matrix", after=curpos)
            curpos += 1
            hdu.header.set("CD1_2", CD1_2, "Coordinate matrix", after=curpos)
            curpos += 1
            hdu.header.set("CD2_1", CD2_1, "Coordinate matrix", after=curpos)
            curpos += 1
            hdu.header.set("CD2_2", CD2_2, "Coordinate matrix", after=curpos)
            curpos += 1
        else:
            hdu.header.set(
                "CD1_1",
                self.focalplane.wcs.cd_1_1[ext_number - 1],
                "Coordinate matrix",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "CD1_2",
                self.focalplane.wcs.cd_1_2[ext_number - 1],
                "Coordinate matrix",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "CD2_1",
                self.focalplane.wcs.cd_2_1[ext_number - 1],
                "Coordinate matrix",
                after=curpos,
            )
            curpos += 1
            hdu.header.set(
                "CD2_2",
                self.focalplane.wcs.cd_2_2[ext_number - 1],
                "Coordinate matrix",
                after=curpos,
            )
            curpos += 1

        return

    def _write_focalplane_keywords(self, ext_number, hdu):
        """
        Define and write the focal plane keywords to the header.
        ext_number starts at 1 (for MEF and FITS).
        """

        curpos = len(hdu.header)
        hdu.header.add_comment(
            "==================================================================",
            after=curpos,
        )
        hdu.header.add_comment("AzCam Focal plane", after=curpos + 1)
        hdu.header.add_comment(
            "==================================================================",
            after=curpos + 2,
        )
        curpos += 3
        hdu.header.set(
            "AMP-CFG",
            self.focalplane.amp_cfg[ext_number - 1],
            "Amplifier configuration",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "DET-NUM",
            self.focalplane.det_number[ext_number - 1],
            "Detector number",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "EXT-NUM",
            self.focalplane.ext_number[ext_number - 1],
            "extension number",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "JPG-EXT",
            self.focalplane.jpg_ext[ext_number - 1],
            "Image section",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "DET-POSX",
            self.focalplane.detpos_x[ext_number - 1],
            "Detector position in X",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "DET-POSY",
            self.focalplane.detpos_y[ext_number - 1],
            "Detector position in Y",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "Ext-PosX",
            self.focalplane.extpos_x[ext_number - 1],
            "Amplifier position in X",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "Ext-PosY",
            self.focalplane.extpos_y[ext_number - 1],
            "Amplifier position in Y",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "AMP-PIX1",
            self.focalplane.amppix1[ext_number - 1],
            "Amplifier pixel position in X",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "AMP-PIX2",
            self.focalplane.amppix2[ext_number - 1],
            "Amplifier pixel position in Y",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "GAIN",
            self.focalplane.gains[ext_number - 1],
            "Amplifier gain",
            after=curpos,
        )
        curpos += 1
        hdu.header.set(
            "RDNOISE",
            self.focalplane.rdnoises[ext_number - 1],
            "Amplifier read noise",
            after=curpos,
        )
        curpos += 1

        return



"""
Contains the ControllerArc class.
"""

import os

import azcam
import azcam.exceptions
from azcam.tools.controller import Controller

from .camera_server import CameraServerInterface


class ControllerArc(Controller):
    """
    The controller class for ARC Gen1, Gen2, and Gen3 controllers.
    """

    def __init__(self, tool_id="controller", description=None):
        super().__init__(tool_id, description)

        self.controller_class = "arc"

        # selected video channel(s) for switched systems
        self.video_select = 0
        # video gain flag
        self.video_gain = 1
        # True to generate fake data instead of real data
        self.synthetic_data = 0

        # True if image data from ControllserServer is binary (not a file)
        self.binary_image_data = 1

        # boards
        self.timing_board = None
        self.clock_boards = [None]
        self.video_boards = [None]
        self.utility_board = None
        # True (1) if the PCI board is installed
        self.pci_board_installed = 1
        # PCI DSP code filename
        self.pci_file = ""
        # True (1) if the timing board is installed
        self.timing_board_installed = 1
        # timing DSP code filename
        self.timing_file = ""
        # True (1) if the utlity board is installed
        self.utility_board_installed = 1
        # utility DSP code filename
        self.utility_file = ""

        # video speed setting
        self.video_speed = 1

        # True to use read lock in ControllerServer
        self.use_read_lock = 0

        # controller server object communicates with ControllerServer
        self.camserver = CameraServerInterface()

        # DSP replies
        self.DON = 0x00444F4E
        self.RDR = 0x00524452
        self.ERR = 0x00455252
        self.SYR = 0x00535952
        self.TIMEOUT = 0x544F5554
        self.READOUT = 0x524F5554
        self.RDA = 0x00524441
        self.IIA = 0x00494941

        # DSP memory locations
        self.Y_CAMSTAT = 0x0  # not used GEN1
        self.Y_NSDATA = 0x1
        self.Y_NPDATA = 0x2
        self.Y_NSBIN = 0x3
        self.Y_NPBIN = 0x4
        self.Y_NSAMPS = 0x5  # not used anymore
        self.Y_NPAMPS = 0x6  # not used anymore
        self.Y_NSCLEAR = 0x7
        self.Y_NPCLEAR = 0x8
        self.Y_NSPRESKIP = 0x9
        self.Y_NSUNDERSCAN = 0xA
        self.Y_NSSKIP = 0xB
        self.Y_NSPOSTSKIP = 0xC
        self.Y_NSOVERSCAN = 0xD
        self.Y_NPPRESKIP = 0xE
        self.Y_NPUNDERSCAN = 0xF
        self.Y_NPSKIP = 0x10
        self.Y_NPPOSTSKIP = 0x11
        self.Y_NPOVERSCAN = 0x12
        self.Y_NPXSHIFT = 0x13
        self.Y_FRAMET = 0x15
        self.Y_PREFLASH = 0x16
        self.Y_NSIMAGE = 0x1B
        self.Y_NPIMAGE = 0x1C
        self.Y_NBOXES = 0x34
        self.Y_NSBIAS = 0x35
        self.Y_NSREAD = 0x36
        self.Y_NPREAD = 0x37
        self.X_STATUS = 0x0
        self.X_OPTIONS = 0x1

        # controller boards
        self.PCIBOARD = 1
        self.TIMINGBOARD = 2
        self.UTILITYBOARD = 3

    def set_boards(self):
        """
        Sets the boards installed in an ARC controller.
        If boards are not specified in this call then controller.timing_board, .clock_boards, .video_boards,
        and .utility_board values are used.

        Timing boards:  'gen1','gen2','arc22'
        Clock boards:   'gen1','gen2','arc32'          (gen1 -> analog combo board)
        Video boards:   'gen1','gen2','arc45','arc48'  (gen1 -> analog combo board)
        Utility boards: 'gen1','gen2','gen3'           (same board but different programming)
        """

        # timing board  (required, controller_type based on this board)
        self.timing_board_installed = 1
        if self.timing_board == "gen1":
            self.controller_type = "gen1"
        elif self.timing_board == "gen2":
            self.controller_type = "gen2"
        elif self.timing_board == "arc22":
            self.controller_type = "gen3"
        else:
            raise azcam.exceptions.AzcamError(
                f"Unrecognized timing board {self.timing_board}"
            )

        # utility board - optional
        if self.utility_board is None:
            self.utility_board_installed = 0
        elif self.utility_board in ["gen1", "gen2", "gen3"]:
            self.utility_board_installed = 1
        else:
            raise azcam.exceptions.AzcamError(
                f"Unrecognized utility board name: {self.utility_board}"
            )

        return

    def initialize(self):
        """
        Initialize controller hardware, loading PCI code as needed.
        """

        if self.is_initialized:
            return

        if not self.is_enabled:
            azcam.exceptions.warning(f"{self.description} is not enabled")
            return

        self.set_boards()

        reply = self.camserver.get("ControllerType")
        if reply[0] == "OK":
            self.is_initialized = True
            if reply[1] == "0" or reply[1] == 0 or reply[1].lower() == "demo":
                self.interface_type = 0
                azcam.exceptions.warning("ControllerServer running in DEMO mode")
                return
            else:
                self.interface_type = int(reply[1])
        elif reply[0] == "DEMO":
            azcam.exceptions.warning("ControllerServer running in DEMO mode")
            self.interface_type = 0
        else:
            raise azcam.exceptions.AzcamError(
                "Could not initialize controller interface"
            )

        # do this only once
        if self.interface_type != 4:  # don't load PCIe DSP code
            azcam.log("Loading PCI file %s" % os.path.basename(self.pci_file))
            self.upload_dsp_file(1, self.pci_file)
            azcam.log("PCI file loaded.")

        return

    # *** reset ***

    def reset_controller(self):
        """
        Issues the ResetController command to the controller.
        """

        # SYR = 0x00535952

        try:
            reply = self.camserver.command("resetcontroller")
            if reply[0] == "ERROR":
                raise azcam.exceptions.AzcamError(f"SYR reset error: {reply[1:]}")
        except azcam.exceptions.AzcamError as e:
            if e.error_code == 1:
                raise azcam.exceptions.AzcamError(
                    "Controller reset error, check power and fibers", error_code=1
                )
            if e.error_code == 2:
                raise azcam.exceptions.AzcamError(
                    "Could not connect to controller server", error_code=2
                )
            else:
                raise

        return

    def reset(self):
        """
        Reset controller using current attributes.
        May warn that the controller could not be reset.
        """

        # reset flag even is system has previously been reset
        self.is_reset = 0

        if not self.is_initialized:
            self.initialize()

        self.reset_controller()

        # restore PCIFILE keyword
        self.set_keyword(
            "PCIFILE",
            os.path.basename(self.pci_file),
            "PCI board DSP code filename",
            "str",
        )

        if self.timing_board_installed:
            azcam.log("Loading timing file %s" % os.path.basename(self.timing_file))
            self.upload_dsp_file(2, self.timing_file)
            azcam.log("Timing board file loaded.")

        if self.utility_board_installed:
            azcam.log("Loading utility file %s" % os.path.basename(self.utility_file))
            self.upload_dsp_file(3, self.utility_file)
            azcam.log("Utility board file loaded.")

        # once code is loaded, controller is "reset"
        self.is_reset = 1

        self.set_bias_voltages()
        self.set_shutter(0)
        self.power_on()
        self.start_idle()
        self.set_video_gain(self.video_gain)
        self.set_video_speed(self.video_speed)
        self.select_video_outputs()
        self.set_roi()
        self.set_exposuretime(0)  # new was .exposure_time

        return

    # *** ROI shifting parameters  ***

    def set_roi(self):
        """
        Sets the ROI parameters values in the controller based on focalplane parameters.
        Sends parameters to the controller.
        """

        # send parameters to controller in order to do all hardware communication here
        if self.is_reset:
            self._write_controller_roi()

        # update ControllerServer for image size
        if self.is_reset:
            self.camserver.set("NumberPixelsImage", self.detpars.numpix_image)

        return

    def _write_controller_roi(self):
        """
        Write clocking parameters to controller.
        """

        # number of total pixels in image for data transfer
        self.write_memory(
            "Y", self.TIMINGBOARD, self.Y_NSIMAGE, self.detpars.numcols_image
        )
        self.write_memory(
            "Y", self.TIMINGBOARD, self.Y_NPIMAGE, self.detpars.numrows_image
        )

        # frame transfer skip size
        self.write_memory("Y", self.TIMINGBOARD, self.Y_FRAMET, self.detpars.framet)

        # number of data pixels to shift
        self.write_memory("Y", self.TIMINGBOARD, self.Y_NSDATA, self.detpars.xdata)
        self.write_memory("Y", self.TIMINGBOARD, self.Y_NPDATA, self.detpars.ydata)

        # set binning
        self.write_memory("Y", self.TIMINGBOARD, self.Y_NSBIN, self.detpars.col_bin)
        self.write_memory("Y", self.TIMINGBOARD, self.Y_NPBIN, self.detpars.row_bin)

        # write number of pixels to flush
        self.write_memory("Y", self.TIMINGBOARD, self.Y_NSCLEAR, self.detpars.xflush)
        self.write_memory("Y", self.TIMINGBOARD, self.Y_NPCLEAR, self.detpars.yflush)

        # write skipping parameters
        self.write_memory(
            "Y", self.TIMINGBOARD, self.Y_NSPRESKIP, self.detpars.xpreskip
        )
        self.write_memory(
            "Y", self.TIMINGBOARD, self.Y_NSUNDERSCAN, self.detpars.xunderscan
        )
        self.write_memory("Y", self.TIMINGBOARD, self.Y_NSSKIP, self.detpars.xskip)
        self.write_memory(
            "Y", self.TIMINGBOARD, self.Y_NSPOSTSKIP, self.detpars.xpostskip
        )
        self.write_memory(
            "Y", self.TIMINGBOARD, self.Y_NSOVERSCAN, self.detpars.xoverscan
        )

        self.write_memory(
            "Y", self.TIMINGBOARD, self.Y_NPPRESKIP, self.detpars.ypreskip
        )
        self.write_memory(
            "Y", self.TIMINGBOARD, self.Y_NPUNDERSCAN, self.detpars.yunderscan
        )
        self.write_memory("Y", self.TIMINGBOARD, self.Y_NPSKIP, self.detpars.yskip)
        self.write_memory(
            "Y", self.TIMINGBOARD, self.Y_NPPOSTSKIP, self.detpars.ypostskip
        )
        self.write_memory(
            "Y", self.TIMINGBOARD, self.Y_NPOVERSCAN, self.detpars.yoverscan
        )

        return

    def set_synthetic_data(self, flag="real"):
        """
        Set controller to create synthetic image data.
        Flag is:
        'real' -> normal, real data
        'fake' or 'synthetic' -> create a synthetic image
        """

        # gen1 not supported
        if self.controller_type == "gen1":
            return

        # timing board X:<STATUS
        value = self.read_memory("X", self.TIMINGBOARD, self.X_STATUS)

        if flag.lower() == "real" or flag.lower() not in [
            "real",
            "fake",
            "synthetic",
        ]:
            flag = 0
        else:
            flag = 1
        # make a synthetic image by setting bit 10 in X:<STATUS
        bits = (value & 0xFFFBFF) | ((flag & 1) << 10)
        self.write_memory("X", self.TIMINGBOARD, self.X_STATUS, bits)

        return

    # *** video ***

    def select_video_outputs(self, video_select=-1):
        """
        Send command to utlity board to select output video channel.
        video_select is video select value.
        """

        if not self.utility_board_installed:
            return

        if video_select == -1:
            video_select = self.video_select

        self.video_select = video_select  # set global, used during exposure

        return self.write_memory("Y", 3, 1, video_select)

    def set_video_gain(self, Gain):
        """
        Set video processor gain.
        For gen2 and ARC45 boards:
        Allowed valid gains are 1,2,5,10
        Also resets video speed to its current value
        For gen1:
        Gain=1 is LOW and Gain=2 is HIGH.
        Gain is video gain value.
        """

        if self.video_boards[0] == "gen1":
            if Gain == 1:
                self.board_command("LGN", self.TIMINGBOARD)
            elif Gain == 2:
                self.board_command("HGN", self.TIMINGBOARD)
            else:
                raise ValueError("Gain must be 1 or 2")

            self.video_gain = Gain
            self.set_keyword("DETGAIN", self.video_gain, "Video gain setting", "int")
            self.set_keyword("VIDGAIN", self.video_gain, "Video gain setting", "int")

        elif self.video_boards[0] in ["gen2", "arc45"]:
            if Gain in [1, 2, 5, 10]:
                speed = self.video_speed - 1
                if speed in [0, 1]:
                    self.board_command("SGN", self.TIMINGBOARD, Gain, speed)
                else:
                    raise azcam.exceptions.AzcamError("Speed must be 1 or 2")
            else:
                raise azcam.exceptions.AzcamError("Gain must be 1, 2, 5, or 10")

            self.video_gain = Gain
            self.set_keyword("DETGAIN", self.video_gain, "Video gain setting", "int")
            self.set_keyword("VIDGAIN", self.video_gain, "Video gain setting", "int")
            self.set_keyword("VIDSPEED", self.video_speed, "Video speed setting", "int")

        return

    def set_video_speed(self, Speed):
        """
        Sets video processor speed for gen2 and arc45 video boards.  Ignored for other boards.
        This command also resets video gain to current value.
        Speed is video speed value (1 slow, 2 fast).
        Speed 0 means no change.
        """

        if Speed not in [0, 1, 2]:
            raise azcam.exceptions.AzcamError("Speed must be 0-2")

        # only used for gen2 and arc45 video boards
        if self.video_boards[0] in ["gen2", "arc45", "sdsu2"]:
            self.video_speed = Speed
            speed = self.video_speed - 1
            self.board_command("SGN", self.TIMINGBOARD, self.video_gain, speed)
            self.set_keyword("VIDSPEED", self.video_speed, "Video speed setting", "int")

        return

    def set_video_offset(self, BoardNumber, DAC, DacValue):
        """
        Sets a video offset value.
        BoardNumber is the controller board number.
        DAC is DAC number.
        DacValue is DAC value to set.
        """

        if self.video_boards[0] == "arc48":  # assume all board types are the same
            self.board_command("SVO", self.TIMINGBOARD, BoardNumber, DAC, DacValue)
        else:
            raise azcam.exceptions.AzcamError(
                "Command set_video_offset not supported for this video board"
            )

        return

    # *** shutter ***

    def set_shutter_state(self, flag: bool = 0):
        """
        Sets the shutter state during an exposure.
        Args:
            flag: open(True) or close(False) shutter during exposure
        """

        # timing board X:<STATUS for gen2/3
        # utility board X:<OPTIONS for gen 1

        if self.controller_type == "gen1":
            reg = self.X_OPTIONS
            board = self.UTILITYBOARD
        else:
            reg = self.X_STATUS
            board = self.TIMINGBOARD

        # note: this read may error with TIMEOUT or other status and we would not know...
        value = self.read_memory("X", board, reg)

        # bit 11 defines shutter state for gen2/3
        # bit 0 for gen1

        if flag:
            flag = 1
        else:
            flag = 0

        if self.controller_type == "gen1":
            bits = (value & 0xFFFFFE) | ((flag & 1) << 0)
        else:
            bits = (value & 0xFFF7FF) | ((flag & 1) << 11)

        self.write_memory("X", board, reg, bits)

        return

    def set_shutter(self, state):
        """
        Open or close controller shutter.
        """

        if self.controller_type == "gen1":
            board = self.UTILITYBOARD
        else:
            board = self.TIMINGBOARD

        if state:
            self.board_command("OSH", board)
        else:
            self.board_command("CSH", board)

        return

    # *** board commands ***

    def board_command(self, Command, BoardNumber, Arg1=-1, Arg2=-1, Arg3=-1, Arg4=-1):
        """
        Send a specific command to an ARC controller board.
        The reply from the board is often 'DON' but could be data.
        Command is the board command to send.
        BoardNumber is controller board number.
        ArgN are arguments for command.
        """

        # change 3 char ascii string to integer
        cmdnum = (ord(Command[0]) << 16) + (ord(Command[1]) << 8) + (ord(Command[2]))

        reply = self.camserver.command(
            "BoardCommand "
            + str(cmdnum)
            + " "
            + str(BoardNumber)
            + " "
            + str(Arg1)
            + " "
            + str(Arg2)
            + " "
            + str(Arg3)
            + " "
            + str(Arg4)
        )

        # check for ERROR
        if reply[0] == "ERROR":
            raise azcam.exceptions.AzcamError(reply[1:][0])

        # check for demo mode: ["DEMO", 0]
        if reply[0] == "DEMO":
            return reply[1]

        # convert controller DSP codes
        try:
            rep = reply[1]
            if rep.startswith("0x"):
                rep = rep.lstrip("0x")
                irep = int(rep, 16)
            else:
                irep = int(rep, 10)
        except Exception:
            return reply[1]

        if irep == self.DON:
            reply1 = "DON"
        elif irep == self.RDR:
            reply1 = "RDR"
        elif irep == self.ERR:
            reply1 = "ERR"
        elif irep == self.SYR:
            reply1 = "SYR"
        elif irep == self.TIMEOUT:
            reply1 = "TIMEOUT"
        elif irep == self.READOUT:
            reply1 = "READOUT"
        elif irep == self.RDA:
            reply1 = "RDA"
        else:
            return reply[1]

        return reply1

    def load_application(self, BoardNumber, ApplicationNumber):
        """
        Load an ARC controller DSP application.
        BoardNumber is controller board number.
        ApplicationNumber is application number to load.
        """

        self.board_command("LDA", BoardNumber, ApplicationNumber)

        return

    def flush(self, Cycles=1):
        """
        Flush or clear out the detector.
        Returns after clearing is finished which could take many seconds.
        """

        for _ in range(Cycles):
            self.board_command("CLR", self.TIMINGBOARD)

        return

    def clear_switches(self):
        """
        Clear ARC controller switches.
        """

        self.board_command("CSW", self.TIMINGBOARD)

        return

    def power_off(self):
        """
        Turn off ARC controller internal power.
        """

        if self.controller_type == "gen1":
            board = self.UTILITYBOARD
        else:
            board = self.TIMINGBOARD

        self.board_command("POF", board)

        return

    def power_on(self):
        """
        Turn on ARC controller internal power.
        """

        if self.controller_type == "gen1":
            board = self.UTILITYBOARD
        else:
            board = self.TIMINGBOARD

        self.board_command("PON", board)

        return

    def parshift(self, NumRows=1):
        """
        Shift CCD detector NumRows rows.
        NumRows is number of rows to shift, positive toward and negative away from amplifier or origin.
        """

        Y_NPXSHIFT = 0x13

        self.write_memory("Y", self.TIMINGBOARD, Y_NPXSHIFT, abs(int(NumRows)))
        if NumRows > 0:
            self.board_command("FPX", self.TIMINGBOARD)
        else:
            self.board_command("RPX", self.TIMINGBOARD)

        return

    def set_bias_number(self, BoardNumber, DAC, Type, DacValue):
        """
        Sets a bias value.
        BoardNumber is the controller board number.
        DAC is DAC number.
        Type is 'VID' or 'CLK'.
        DacValue is DAC value for voltage.
        """

        if self.video_boards[0] == "gen1":
            raise azcam.exceptions.AzcamError(
                "Command set_bias_number not supported for this controller"
            )
        elif self.video_boards[0] in [
            "arc48",
            "arc47",
        ]:  # assume all board types are the same
            self.board_command(
                "SBN", self.TIMINGBOARD, BoardNumber, Type, DAC, DacValue
            )
        else:
            self.board_command(
                "SBN", self.TIMINGBOARD, BoardNumber, DAC, Type, DacValue
            )

        return

    def set_bias_voltages(self):
        """
        Turns on all DC biases.
        """

        self.board_command("SBV", self.TIMINGBOARD)

        return

    def set_dc_mode(self, Mode):
        """
        Sets DC mode.
        Mode is value to set.
        """

        self.board_command("SDC", self.TIMINGBOARD, Mode)

        return

    def set_mux(self, BoardNumber, MUX1, MUX2):
        """
        Sets MUX values.
        BoardNumber is controller board number.
        MUX1 is MUX 1 value.
        MUX2 us MUX 2 value.
        """

        self.board_command("SMX", self.TIMINGBOARD, BoardNumber, MUX1, MUX2)

        return

    def start_idle(self):
        """
        Start idle clocking.
        """

        self.board_command("IDL", self.TIMINGBOARD)

        return

    def stop_idle(self):
        """
        Stop idle clocking.
        """

        self.board_command("STP", self.TIMINGBOARD)

        return

    def write_control(self, Word):
        """
        Writes the Control Word.
        Word is the data to write.
        """

        self.board_command("WRC", self.TIMINGBOARD, Word)

        return

    def write_memory(self, Type, BoardNumber, Address, value):
        """
        Write a word to a DSP memory location.
        Type is P, X, Y, or R memory space.
        BoardNumber is controller board number.
        Address is memory address to write.
        value is data to write.
        """

        if Type == "P":
            arg = 0x100000
        elif Type == "X":
            arg = 0x200000
        elif Type == "Y":
            arg = 0x400000
        elif Type == "R":
            arg = 0x800000
        else:
            raise azcam.exceptions.AzcamError("Invalue write_memory type")

        self.board_command("WRM", BoardNumber, arg | Address, value)

        return

    def read_memory(self, Type, BoardNumber, Address):
        """
        Read from DSP memory.
        Type is P, X, Y, or R memory space.
        BoardNumber is controller board number.
        Address is memory address to read.
        """

        if Type == "P":
            arg = 0x100000
        elif Type == "X":
            arg = 0x200000
        elif Type == "Y":
            arg = 0x400000
        elif Type == "R":
            arg = 0x800000
        else:
            raise azcam.exceptions.AzcamError("Invalue Type")

        BoardNumber = int(BoardNumber)
        Address = int(Address)

        reply = self.board_command("RDM", BoardNumber, arg | Address)

        return int(reply)

    # *** DSP files ***

    def upload_dsp_file(self, BoardNumber, filename):
        """
        Sends DSP a file to a controllercontaining DSP code to the PCI, timing, or utility boards.
        """

        if filename == "":
            return

        csfile = self.upload_file(filename)

        self.load_file(BoardNumber, csfile)

        # set keyword for file loaded
        if BoardNumber == 1:
            self.set_keyword(
                "PCIFILE",
                os.path.basename(filename),
                "PCI board DSP code filename",
                "str",
            )
        elif BoardNumber == 2:
            self.set_keyword(
                "TIMFILE",
                os.path.basename(filename),
                "Timing board DSP code filename",
                "str",
            )
        elif BoardNumber == 3:
            self.set_keyword(
                "UTILFILE",
                os.path.basename(filename),
                "Utility board DSP code filename",
                "str",
            )

        return

    def load_file(self, BoardNumber, filename):
        """
        Write a file containing DSP code to the PCI, timing, or utility boards.
        The file must be on the ControllerServer file system.
        BoardNumber is controller board number.
        filename is file to load (.lod type).
        """

        if BoardNumber == 1:
            filename = os.path.normpath(filename)
            self.camserver.load_file(1, filename)
        elif BoardNumber == 2:
            filename = os.path.normpath(filename)
            self.camserver.load_file(2, filename)
        elif BoardNumber == 3:
            filename = os.path.normpath(filename)
            self.camserver.load_file(3, filename)
        else:
            raise azcam.exceptions.AzcamError("Invalid board number")

        return

    # *** files ***

    def upload_file(self, filename):
        """
        Sends a local file to the controller server to be written to its file system.
        Returns uploaded filename on controller server.
        """

        with open(filename, "r") as f:
            fbuffer = f.read()

        # send file as binary, reply is filename on controller server
        reply = self.camserver.upload_file(fbuffer)

        return reply

    # *** readout ***

    def start_readout(self):
        """
        Start readout immediately.
        Returns imemdiately, does not wait for readout to finish.
        """

        return self.read_image()

    def get_pixels_remaining(self):
        """
        Return number of remaining pixels to be read (counts down).
        """

        reply = self.camserver.get("PixelCount")
        count = int(reply[1])
        return max(0, self.detpars.numpix_image - count)

    # *** exposure time ***

    def set_exposuretime(self, ExposureTime: float):
        """
        Write the exposure time (in seconds) to the controller.
        """

        ExposureTime = float(ExposureTime)
        self.exposure_time = ExposureTime  # new

        et_msec = int(ExposureTime * 1000)

        self.camserver.set("ExposureTime", et_msec)

        return

    def get_exposuretime(self):
        """
        Return the exposure time from the controller (in seconds).
        """

        reply = self.camserver.get("ExposureTime")  # this in msec

        return float(reply[1]) / 1000.0

    # *** misc ***

    def start_exposure(self):
        """
        Start exposure (integration).
        Returns immediately, not waiting for exposure to finish.
        """

        self.camserver.command("StartExposure")

        return

    def ioctl(self, Command):
        """
        Issue a low-level ARC controller IO command on the controller server.
        """

        return self.camserver.command("ioctl " + str(Command))

    def exposure_abort(self):
        """
        Abort an integration.
        """
        # reply=self.Boardcommand('PEX',2)               # test 16Nov12
        # reply=self.Boardcommand('AEX',2)               # test 16Nov12
        # return reply

        return self.camserver.command("AbortExposure")

    def readout_abort(self):
        """
        Abort a readout in progress.
        """

        return self.camserver.command("AbortReadout")

    def exposure_pause(self):
        """
        Pause an integration which is in progress.
        """

        return self.camserver.command("PauseExposure")

    def read_image(self):
        """
        Start readout of detector.
        """

        return self.camserver.command("ReadImage")

    def exposure_resume(self):
        """
        Resume a paused integration.
        """

        return self.camserver.command("ResumeExposure")

    def update_exposuretime_remaining(self):
        """
        Return remaining exposure time (in seconds).
        """

        reply = self.camserver.get("ExposureTimeRemaining")  # returns elapsed

        elapsed = int(reply[1])  # milliseconds, was + 1

        if elapsed < 0:
            elapsed = 0

        return max(0, self.exposure_time * 1000 - elapsed) / 1000.0

    # **************************************************************************
    # Test Commands
    # **************************************************************************

    def test_datalink(self, board_number=0, value="counter", loops=10):
        """
        Test comminications to one or more controller boards.
        BoardNumber is the board number (0=all boards, 1=PCI, 2=Timing, 3=Utility).
        value is an integer.
        Loops is the number of times to repeat command.
        """

        board_number = int(board_number)
        loops = int(loops)

        bmin = 1

        if board_number == 0:
            bmax = 4 if self.utility_board_installed else 3
        else:
            bmin = board_number
            bmax = board_number + 1

        for board in range(bmin, bmax):
            for loop in range(loops):
                value = loop if value == "counter" else value
                reply = self.board_command("TDL", board, value)
                if int(reply) == value:
                    continue
                else:
                    raise azcam.exceptions.AzcamError(
                        f"Communication to board {board} failed on loop {loop}"
                    )

        return



"""
Contains the ExposureArc class.
"""

import os
import threading
import time

import azcam
import azcam.exceptions
from azcam.tools.exposure import Exposure

from .receive_data import ReceiveData


class ExposureArc(Exposure):
    """
    Defines the exposure class for ARC controllers which makes an exposure.
    """

    def __init__(self, tool_id="exposure", description=None):
        super().__init__(tool_id, description)

        self.receive_data = ReceiveData(self)

    def integrate(self):
        """
        Integration.
        """

        # start integration
        self.exposure_flag = self.exposureflags["EXPOSING"]
        imagetype = self.image_type.lower()

        # start exposure
        if imagetype != "zero":
            azcam.log("Integration started")
        azcam.db.tools["controller"].start_exposure()
        self.dark_time_start = time.time()
        reply = self.get_exposuretime_remaining()
        remtime = reply
        lasttime = remtime

        # countdown and check for async. ExposureFlag changes
        loopcount = 0

        while remtime > 0.6:
            if self.exposure_flag == self.exposureflags["EXPOSING"]:  # no EF changes
                time.sleep(min(remtime, 0.5))
                reply = self.get_exposuretime_remaining()
                remtime = reply
                azcam.log(f"Integration: {remtime:0.3f} seconds remaining", level=3)
                if remtime == lasttime:
                    loopcount += 1
                else:
                    loopcount = 0
                    lasttime = remtime

            if loopcount > 20:
                azcam.log("ERROR Integration time stuck")
                self.exposure_flag = self.exposureflags["ABORT"]
                azcam.db.tools["controller"].exposure_abort()
                break
            elif (
                self.exposure_flag == self.exposureflags["ABORT"]
            ):  # AbortExposure received
                if self.is_exposure_sequence:
                    azcam.log("Stopping exposure sequence")
                    self.is_exposure_sequence = 0
                    self.exposure_sequence_number = 1
                    self.exposure_flag = self.exposureflags["EXPOSING"]
                else:
                    azcam.db.tools["controller"].exposure_abort()
                    break
            elif (
                self.exposure_flag == self.exposureflags["PAUSE"]
            ):  # PauseExposure received
                azcam.db.tools["controller"].exposure_pause()
                self.exposure_flag = self.exposureflags["PAUSED"]
                azcam.log("Integration paused")
            elif (
                self.exposure_flag == self.exposureflags["RESUME"]
            ):  # ResumeExposure received
                azcam.db.tools["controller"].exposure_resume()
                self.exposure_flag = self.exposureflags["EXPOSING"]
                reply = self.get_exposuretime_remaining()
                remtime = reply
                azcam.log("Integration resumed")
            elif (
                self.exposure_flag == self.exposureflags["READ"]
            ):  # ReadExposure received
                remtime = 0.0
                self.exposure_time_actual = (
                    self.exposure_time - self.exposure_time_remaining
                )
                break
            elif (
                self.exposure_flag == self.exposureflags["PAUSE"]
            ):  # already paused so just loop
                time.sleep(0.5)

        if self.exposure_flag == self.exposureflags["ABORT"]:  # abort in remaining time
            azcam.log("Integration aborted")
        else:
            time.sleep(remtime + 0.1)
            self.exposure_flag = self.exposureflags["READ"]  # set to readout

        self.dark_time = time.time() - self.dark_time_start

        # turn off comp lamps
        if not self.comp_sequence:
            if self.comp_exposure:
                if not azcam.db.tools["instrument"].shutter_strobe:
                    azcam.db.tools["instrument"].comps_off()
                azcam.db.tools["instrument"].set_comps("shutter")

        # extra close shutter command
        reply = azcam.db.tools["controller"].set_shutter(0)

        # set times
        self.exposure_time_remaining = 0
        if imagetype == "zero":
            self.exposure_time = self.exposure_time_saved

        if self.exposure_flag == self.exposureflags["ABORT"]:
            azcam.exceptions.warning("Integration aborted")
        else:
            azcam.log("Integration finished", level=2)

        return

    def readout(self):
        """
        Exposure readout.
        """

        self.exposure_flag = self.exposureflags["READ"]

        imagetype = self.image_type.lower()

        if imagetype == "ramp":
            azcam.db.tools["controller"].set_shutter(1)

        if self.tdi_mode:
            self.set_tdi_delay(True)

        # start readout
        azcam.db.tools["controller"].start_readout()
        self.exposure_flag = self.exposureflags["READOUT"]
        azcam.log("Readout started")

        # start data transfer, returns when all data is received
        try:
            self.receive_data.receive_image_data(self.image.focalplane.numpix_image * 2)
        except azcam.exceptions.AzcamError as e:
            if e.error_code == 3:
                azcam.log("Exposure aborted")
            else:
                raise

        # check if aborted by user
        if azcam.db.abortflag and self.is_exposure_sequence:  # stop exposure sequence
            azcam.log("User abort in exposure sequence")

        self.image.valid = 1

        if imagetype == "ramp":
            azcam.db.tools["controller"].set_shutter(0)

        if self.tdi_mode:
            self.set_tdi_delay(False)

        if self.exposure_flag == self.exposureflags["ABORT"]:
            # if aborted in a sequence, reset flags and let finish without error
            if self.is_exposure_sequence:
                azcam.log("Stopping exposure sequence in Exposure")
                self.is_exposure_sequence = 0
                self.exposure_sequence_number = 1
                self.exposure_flag = self.exposureflags["READ"]
            else:
                azcam.log("Readout aborted")
                raise azcam.exceptions.AzcamError("Readout aborted", error_code=3)
        elif self.exposure_flag == self.exposureflags["ERROR"]:
            if self.is_exposure_sequence:
                self.exposure_flag = self.exposureflags["NONE"]
                raise azcam.exceptions.AzcamError("Exposure sequence error occurred")
            else:
                self.exposure_flag = self.exposureflags["ABORT"]
                azcam.log("Readout aborted")
                raise azcam.exceptions.AzcamError("Readout aborted", error_code=3)
        else:
            azcam.log("Readout finished", level=2)
            self.exposure_flag = self.exposureflags["NONE"]

        return

    def end(self):
        """
        Completes an exposure by writing file and displaying image.
        """

        self.exposure_flag = self.exposureflags["WRITING"]

        # if remote write, LocalFile is local temp file
        if self.send_image:
            LocalFile = self.temp_image_file + "." + self.get_extname(self.filetype)
            try:
                os.remove(LocalFile)
            except FileNotFoundError:
                pass
        else:
            LocalFile = self.get_filename()
        self.last_filename = LocalFile

        # wait for image data to be received
        loop = 0
        while not self.image.valid and loop < 100:
            loop += 1
            time.sleep(0.050)
            if loop >= 100:
                azcam.log("ERROR image data not received in time")

        # update controller header with keywords which might have changed
        et = float(int(self.exposure_time_actual * 1000.0) / 1000.0)
        dt = float(int(self.dark_time * 1000.0) / 1000.0)
        azcam.db.headers["exposure"].set_keyword(
            "EXPTIME", et, "Exposure time (seconds)", "float"
        )
        azcam.db.headers["exposure"].set_keyword(
            "DARKTIME", dt, "Dark time (seconds)", "float"
        )

        # write file(s) to disk
        if self.save_file:
            azcam.log("Writing %s" % LocalFile)

            # write the file to disk
            self.image.overwrite = self.overwrite
            self.image.test_image = self.test_image
            self.image.write_file(LocalFile, self.filetype)
            azcam.log("Writing finished", level=2)

            # set flag that image now written to disk
            self.image.is_written = 1

            # send image to guider software
            if self.guide_mode:
                self.sendimage.send_image(LocalFile)

            # send image to remote image server
            elif self.send_image:
                if self.write_async:
                    self.exposure_flag = self.exposureflags[
                        "NONE"
                    ]  # reset flag now so next exposure can start
                    azcam.log("Sending image asynchronously")
                    sendthread = threading.Thread(
                        target=self.sendimage.send_image,
                        name="writeasync",
                        args=(self.get_filename()),
                    )
                    sendthread.start()

                    # increment file sequence number now and return
                    self.increment_filenumber()
                    self.exposure_flag = self.exposureflags["NONE"]
                    return

                else:
                    azcam.log("Sending image")
                    self.sendimage.send_image(LocalFile, self.get_filename())

        # image data and file are now ready
        self.image.toggle = 1

        # added for sequence pause and readout issue
        self.exposure_time = self.exposure_time_saved
        self.set_exposuretime(self.exposure_time)

        # reset idle if no flush
        if not self.flush_array:
            azcam.db.tools["controller"].start_idle()

        # display image
        if self.display_image and not self.write_async:
            azcam.log("Displaying image")
            azcam.db.tools["display"].display(self.image)

        # increment file sequence number if image was written
        if self.save_file:
            self.increment_filenumber()

        self.exposure_flag = self.exposureflags["NONE"]

        return

    def abort(self):
        """
        Abort an exposure in progress.
        Really sets the ExposureFlag which is read in during the exposure.
        """

        if (
            azcam.db.tools["controller"].controller_type == "gen2"
            and self.exposure_flag == self.exposureflags["READ"]
        ):
            raise azcam.exceptions.AzcamError(
                "Abort readout not supported for this controller"
            )

        if (
            azcam.db.tools["controller"].controller_type == "gen1"
            and self.exposure_flag == self.exposureflags["READ"]
        ):
            raise azcam.exceptions.AzcamError(
                "Abort readout not supported for this controller"
            )

        if self.exposure_flag != self.exposureflags["NONE"]:
            self.exposure_flag = self.exposureflags["ABORT"]

        return

    # **********************************************************************************************
    # TDI commands
    # **********************************************************************************************

    def set_tdi_delay(self, Flag):
        """
        Set TDI line delay On or Off.
        Flag==True sets delay to self.tdi_delay.
        Flag==Flase sets line delay to normal value, self.par_delay.
        """

        if Flag:
            Delay = self.tdi_delay
            azcam.db.tools["controller"].set_keyword(
                "TdiDelay", Delay, "TDI delay multiplier", "int"
            )
        else:
            Delay = self.par_delay
            azcam.db.tools["controller"].delete_keyword("TdiDelay")

        azcam.db.tools["controller"].write_memory(
            "Y", azcam.db.tools["controller"].TIMINGBOARD, 0x20, Delay
        )

        return



from dash import Dash, html, dcc, callback, Input, Output, State, ctx
import dash_bootstrap_components as dbc

import azcam


def tabs_buttons():

    buttons = dbc.ButtonGroup(
        [
            dbc.Button("Exposure", id="exposure_tab_btn", className="me-1", n_clicks=0),
            dbc.Button("Filename", id="filename_tab_btn", className="me-1", n_clicks=0),
            dbc.Button("Detector", id="detector_tab_btn", className="me-1", n_clicks=0),
            dbc.Button("Options", id="options_tab_btn", className="me-1", n_clicks=0),
        ],
        vertical=False,
    )

    @callback(
        Output("tabs", "active_tab", allow_duplicate=False),
        Output("image_type_dropdown", "value"),
        Output("title_input", "value"),
        Output("test_image", "value"),
        Output("et", "value"),
        Output("seq_num", "value"),
        Output("seq_delay", "value"),
        Input("exposure_tab_btn", "n_clicks"),
        prevent_initial_call=False,
    )
    def on_button_click_exposure_tab_btn(n):
        exposure = azcam.db.tools["exposure"]
        image_type_dropdown = exposure.get_image_type()
        title_input = exposure.get_image_title()
        test_image = exposure.test_image
        et = exposure.exposure_time
        seq_num = exposure.exposure_sequence_total
        seq_delay = exposure.exposure_sequence_delay
        return [
            "exposure_tab",
            image_type_dropdown,
            title_input,
            test_image,
            et,
            seq_num,
            seq_delay,
        ]

    @callback(
        Output("tabs", "active_tab", allow_duplicate=True),
        Output("folderselect", "value"),
        Output("rootselect", "value"),
        Output("seqnumselect", "value"),
        Input("filename_tab_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_filename_tab_btn(n):
        exposure = azcam.db.tools["exposure"]
        folder = exposure.folder
        seq = exposure.sequence_number
        root = exposure.root

        return ["filename_tab", folder, root, seq]

    @callback(
        Output("tabs", "active_tab", allow_duplicate=True),
        Output("first_col", "value"),
        Output("last_col", "value"),
        Output("col_bin", "value"),
        Output("first_row", "value"),
        Output("last_row", "value"),
        Output("row_bin", "value"),
        Input("detector_tab_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_detector_tab_btn(n):
        firstcol, lastcol, firstrow, lastrow, colbin, rowbin = azcam.db.tools[
            "exposure"
        ].get_roi()

        return [
            "detector_tab",
            firstcol,
            lastcol,
            colbin,
            firstrow,
            lastrow,
            rowbin,
        ]

    @callback(
        Output("tabs", "active_tab", allow_duplicate=True),
        Output("flush_sensor", "value"),
        Output("display_image", "value"),
        Output("enable_instrument", "value"),
        Output("enable_telescope", "value"),
        Output("auto_title", "value"),
        Input("options_tab_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_options_tab_btn(n):
        flush_sensor = azcam.db.tools["exposure"].flush_array
        display_image = azcam.db.tools["exposure"].display_image
        if azcam.db.tools.get("instrument"):
            enable_instrument = azcam.db.tools.get("instrument").is_enabled
        else:
            enable_instrument = 0
        if azcam.db.tools.get("telescope"):
            enable_telescope = azcam.db.tools["telescope"].is_enabled
        else:
            enable_telescope = 0
        auto_title = azcam.db.tools["exposure"].auto_title

        return [
            "options_tab",
            flush_sensor,
            display_image,
            enable_instrument,
            enable_telescope,
            auto_title,
        ]

    return buttons


def tabs(parent):

    exposure_tab = [
        dbc.CardBody(
            [
                parent.exposure_card,
                parent.status_card,
            ]
        )
    ]

    filename_tab = [
        parent.filename_card,
        html.Div(id="filename_out"),
    ]

    detector_tab = [
        parent.detector_card,
    ]

    options_tab = [
        parent.options_card,
    ]

    tabs = dbc.Tabs(
        [
            dbc.Tab(exposure_tab, tab_id="exposure_tab"),
            dbc.Tab(filename_tab, tab_id="filename_tab"),
            dbc.Tab(detector_tab, tab_id="detector_tab"),
            dbc.Tab(options_tab, tab_id="options_tab"),
        ],
        id="tabs",
        active_tab="exposure_tab",
    )

    return tabs



"""
Example logging server which receives and prints azcam log records.
"""

import socketserver
import pickle
import struct
import logging
import ctypes

from loguru import logger


class LoggingStreamHandler(socketserver.StreamRequestHandler):
    def handle(self):
        while True:
            try:
                chunk = self.connection.recv(4)
            except ConnectionResetError:
                break
            if len(chunk) < 4:
                break
            slen = struct.unpack(">L", chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            record = pickle.loads(chunk)
            lrec = logging.makeLogRecord(record)
            outstring = lrec.msg
            # outstring = outstring.replace("| azcam.logger:log:68 - ", "")
            # outstring = outstring.replace(" | INFO   ", "")
            outstring = outstring[65:]

            print(outstring)
            # level, message = record["levelname"], record["msg"]
            # print(message)
            # logger.patch(lambda record: record.update(record)).log(level, message)


def start_and_serve_tcp(port: int = 2404):
    """
    Start logging server.
    """

    # optionally set window title
    try:
        ctypes.windll.kernel32.SetConsoleTitleW("azcamlogger")
    except Exception:
        pass
    print(f"Logging server started on port {port}")
    logging_server = socketserver.TCPServer(("localhost", port), LoggingStreamHandler)
    logging_server.serve_forever()


if __name__ == "__main__":
    start_and_serve_tcp()



"""
Contains the base Instrument class.
"""

import time
from typing import Any, List, Optional

import azcam
from azcam.header import Header, ObjectHeaderMethods
from azcam.tools.tools import Tools


class Instrument(Tools, ObjectHeaderMethods):
    """
    The base instrument tool.
    Usually implemented as the "instrument" tool.
    """

    def __init__(self, tool_id="instrument", description=None):
        Tools.__init__(self, tool_id, description)

        # active comparisons
        self.active_comps = []

        # True if shutter controls comps
        self.shutter_strobe = 0

        # focus position
        self.focus_position = 0

        # system pressures
        self.pressure_ids = [0]

        # system currents
        self.current_ids = [0]

        azcam.db.tools_init["instrument"] = self
        azcam.db.tools_reset["instrument"] = self

        # instrument header object
        self.header = Header("Instrument")
        self.header.set_header("instrument", 3)

        return

    def initialize(self):
        """
        Initialize tool.
        """

        self.is_initialized = 1

        return

    def reset(self):
        """
        Reset tool.
        """

        self.is_reset = 1

        return

    # ***************************************************************************
    # exposure
    # ***************************************************************************

    def exposure_start(self):
        """
        Optional call before exposure starts.
        """

        return

    def exposure_finish(self):
        """
        Optional tempcon.call after exposure finishes.
        """

        return

    # ***************************************************************************
    # comparisons
    # ***************************************************************************
    def get_all_comps(self):
        """
        Return all valid comparison names.
        Useful for clients to determine which type of comparison exposures are supported.
        """

        raise NotImplementedError

    def get_comps(self):
        """
        Return a list of the active comparison lamps.
        """

        return self.active_comps

    def set_comps(self, comp_names: List[str] = None):
        """
        Set comparisons to be turned on and off with comps_on() and comps_off().
        Args:
            comp_names: list of string (or a single string) of comparison names
        """

        if type(comp_names) == str:
            comp_names = comp_names.split(" ")

        self.active_comps = comp_names

        return

    def comps_on(self):
        """
        Turn on active comparisons.
        """

        return

    def comps_off(self):
        """
        Turn off active comparisons.
        """

        return

    def comps_delay(self, delay_time=0):
        """
        Delays for delay_time for comparison lamp warmup.
        May internally set delay based on comp selection.
        Args:
            delay_time: delay in seconds
        """

        time.sleep(delay_time)

        return

    # ***************************************************************************
    # filters
    # ***************************************************************************
    def get_filters(self, filter_id=0):
        """
        Return a list of all available/loaded filters.
        Args:
            filter_id: filter mechanism ID
        """

        raise NotImplementedError

    def get_filter(self, filter_id=0):
        """
        Return the current/loaded filter, typically the filter in the beam.
        Args:
            filter_id: filter mechanism ID
        """

        raise NotImplementedError

    def set_filter(self, filter_name, filter_id=0):
        """
        Set the current/loaded filter, typically the filter in the beam.
        Args:
            filter_name: filter name to set. Could be a number or filter name.
            filter_id: filter mechanism ID
        """

        raise NotImplementedError

    # ***************************************************************************
    # wavelengths
    # ***************************************************************************
    def get_wavelengths(self, wavelength_id: int = 0):
        """
        Returns a list of valid wavelengths.
        Used for filter and LED based systems.
        Args:
            wavelength_id: wavelength mechanism ID
        """

        raise NotImplementedError

    def get_wavelength(self, wavelength_id: int = 0):
        """
        Returns the current wavelength.
        Args:
            wavelength_id: wavelength mechanism ID
        """

        raise NotImplementedError

    def set_wavelength(self, wavelength: Any, wavelength_id: int = 0):
        """
        Set the current wavelength, typically for a filter or grating.
        Args:
            wavelength: wavelength value to set. Could be a number or filter name.
            wavelength_id: wavelength mechanism ID
        """

        raise NotImplementedError

    # ***************************************************************************
    # focus
    # ***************************************************************************
    def get_focus(self, focus_id=0):
        """
        Return the current instrument focus position.
        focus_id is the focus mechanism ID.
        """

        raise NotImplementedError("get_focus")

    def set_focus(self, focus_position, focus_id=0, focus_type="absolute"):
        """
        Move (or step) the instrument focus.
        focus_position is the focus position or step size.
        focus_id is the focus mechanism ID.
        focus_type is "absolute" or "step".
        """

        raise NotImplementedError("set_focus")

    # ***************************************************************************
    # pressure
    # ***************************************************************************
    def get_pressures(self):
        """
        Return a list of all instrument pressures.
        """

        pressures = []
        for pressure_id in self.pressure_ids:
            p = self.get_pressure(pressure_id)
            pressures.append(p)

        return pressures

    def get_pressure(self, pressure_id=0):
        """
        Read an instrument pressure.
        """

        raise NotImplementedError

    def set_pressure(self, pressure, pressure_id=0):
        """
        Sets an instrument pressure.
        """

        raise NotImplementedError

    # ***************************************************************************
    # shutter
    # ***************************************************************************
    def set_shutter(self, state, shutter_id=0):
        """
        Open or close the instrument shutter.
        state is 1 for open and 0 for close.
        shutter_id is the shutter mechanism ID.
        """

        raise NotImplementedError

    # ***************************************************************************
    # power meter
    # ***************************************************************************
    def get_power(self, wavelength: float, power_id: int = 0) -> float:
        """
        Returns power meter reading.
        Args:
            wavelength: wavelength for power meter
            power_id: power ID flag
        Returns:
            mean_power: mean power in Watts/cm2
        """

        raise NotImplementedError

    # ***************************************************************************
    # electrometer
    # ***************************************************************************
    def get_currents(self):
        """
        Return a list of all instrument currents.
        """

        currents = []
        for current_id in self.current_ids:
            p = self.get_current(current_id)
            currents.append(p)

        return currents

    def get_current(self, shutter_state: int = 1, current_id: int = 0) -> float:
        """
        Read instrument current, usually from an electrometer.
        Args:
            current_id: current source ID
            shutter_state: shutter state during read
        Returns:
            current: measured current in amps
        """

        raise NotImplementedError



import time


class ObsTime(object):
    """
    Defines the ObsTime class.
    Used by the exposure tool.
    """

    def __init__(self):

        # UT
        self.utc = []
        # UT date only
        self.date = []
        # UT time only
        self.ut = []
        # local time
        self.local_time = []
        # local time zone
        self.time_zone = []
        # time system
        self.time_system = []

        return

    def update(self, index=0):
        """
        Save the current time parameters into the specified internal index.
        Index is 0 for open shutter time.
        """

        if len(self.utc) < index + 1:
            self.utc.append("")
            self.date.append("")
            self.ut.append("")
            self.local_time.append("")
            self.time_zone.append("")
            self.time_system.append("")

        # get current time and date
        gmt = time.gmtime()
        self.utc[index] = str(gmt[0]) + time.strftime("-%m-%d", gmt)

        self.date[index] = self.utc[index]

        t = time.time()
        fracsec = int(1000 * round(t - int(t), 3))
        self.ut[index] = time.strftime("%H:%M:%S.", gmt) + "%.3u" % fracsec

        self.time_zone[index] = time.timezone / 3600

        self.local_time[index] = time.strftime("%H:%M:%S", time.localtime())

        self.time_system[index] = "UTC"

        return



"""
Contains the CameraServerInterface class for Magelllan controllers.
"""

import azcam
import azcam.exceptions
import azcam.sockets


class CameraServerInterface(object):
    """
    The azcam controller server interface class for Magellan controllers.
    """

    def __init__(self, host: str = "", port: int = 0) -> bool:

        self.host = ""
        self.port = 0

        self.socketserver = azcam.sockets.SocketInterface(host, port)

        self.demo_mode = 0

    def set_server(self, host: str, port: int = 2405) -> None:
        """
        Set host and port of controller server.
        """

        self.host = host
        self.port = port

        self.socketserver.host = host
        self.socketserver.port = port

        return

    def command(self, command: str, terminator: str = "\n"):
        """
        Command method for controller server.
        """

        if self.demo_mode:
            reply = ["DEMO", 0]
        else:
            try:
                reply = self.socketserver.command(command, terminator)
                return reply
            except Exception as e:
                if e.error_code == 2:
                    raise azcam.exceptions.AzcamError("Could not connect to camserver")

    def test(self):
        """
        Echo a message string from controller server.
        """

        return self.echo()

    def echo(self, Message="This is a ControllerServer test message."):
        """
        Echo a message string from controller server.
        """

        return self.command('Echo "' + str(Message) + '"')

    def upload_file(self, fbuffer):
        """
        Sends a file as a binary buffer to the ControllerServer to be written to its file system.
        Returns the name of the file on the ControllerServer file system.
        """

        # send size
        size = len(fbuffer)
        self.command("cmd UploadFile " + str(size))

        # send file buffer
        if not self.demo_mode:
            self.socketserver.send(fbuffer)
            reply = self.socketserver.recv()
            csfilename = reply.split(" ")[1]
        else:
            csfilename = ""

        return csfilename.strip()

    def load_file(self, board, filename):
        """
        Send a a DSP code file to the controller server.
        """

        try:
            reply = self.command("LoadFile " + str(board) + " " + filename)
        finally:
            self.delete_file(filename)  # try and delete file even if error

        return

    def delete_file(self, filename):
        """
        Delete a file which wa uploaded to the controller server.
        """

        return self.command("cmd DeleteFile " + filename)

    def set(self, Parameter, value):
        """
        Set a parameter in the controller server.
        """

        return self.command("Set " + Parameter + " " + str(value))

    def get(self, Parameter):
        """
        Return a paramater from the controller server, as a string.
        """

        if self.demo_mode:
            if Parameter == "ExposureTimeRemaining":
                parval = azcam.db.tools["exposure"].exposure_time * 1000
                reply = ["OK", parval]
            else:
                reply = ["OK", 0]
        else:
            reply = self.command("Get " + Parameter)

        return reply

    def close_server(self):
        """
        Closes the ControllerServer.
        """

        return self.command("CloseServer")

    def restart_server(self):
        """
        Restarts the ControllerServer.
        """

        self.command("RestartServer")
        if not self.demo_mode:
            self.socketserver.close()  # close socket as it is reset in CS

        return

    def reset_server(self):
        """
        Resets the ControllerServer.
        """

        self.command("ResetServer")
        if not self.demo_mode:
            self.socketserver.close()  # close socket as it is reset in CS

        return



"""
DASH web server for azcam
"""

import logging

from dash import Dash, html, dcc, callback, Input, Output, State, ctx
import dash_bootstrap_components as dbc
import dash_daq as daq

import azcam
from azcam.web.status.status_card import statusweb_card


class StatusWeb(object):
    """
    Web server using Dash (plotly).
    """

    def __init__(self) -> None:

        azcam.db.statusweb = self

        self.app = Dash(
            __name__,
            external_stylesheets=[dbc.themes.BOOTSTRAP],
            suppress_callback_exceptions=False,
            requests_pathname_prefix="/status/",
        )
        logging.getLogger("werkzeug").setLevel(logging.CRITICAL)  # stop messages

        self.statusweb_card = statusweb_card()

        # app layout
        self.app.layout = html.Div(
            [
                self.statusweb_card,
                dcc.Interval("statusweb_interval", interval=1_000, n_intervals=0),
            ]
        )

        return

    def set_message(self, message: str = "") -> None:
        """
        Set the browser message.
        """

        azcam.db.tools["exposure"].message = repr(message)
        return



import time

import azcam


def execute():
    """
    Execute queue dataframe.
    """

    df = azcam.db.queue_df

    for irow in df.index:
        print(f"Row {irow}")
        cmd = df.iloc[irow]["Command"]

        if cmd.startswith("print"):
            text = cmd.lstrip("print ")
            print(f" {text}")

        elif cmd.startswith("delay"):
            delay = float(cmd.lstrip("delay "))
            print(f" Delay {delay} secs")
            time.sleep(delay)

        elif cmd.startswith("movefilter"):
            filt = cmd.lstrip("movefilter ")
            print(f" Moving filter to {filt}")

        elif cmd.startswith("steptel"):
            ra_dec = cmd.lstrip("steptel ")
            print(f" Stepping telescope to {ra_dec}")

        elif cmd.startswith("movetel"):
            ra_dec = cmd.lstrip("movetel ")
            print(f" Moving telescope to {ra_dec}")

        time.sleep(0.5)

    return



"""
Contains the base Controller and DetPars classes.
"""

import azcam
from azcam.header import Header, ObjectHeaderMethods
from azcam.tools.tools import Tools


class Controller(Tools, ObjectHeaderMethods):
    """
    Base class for controller tool.
    Usually implemented as the "controller" tool.
    """

    def __init__(self, tool_id: str = "controller", description: str or None = None):
        """
        Args:
            tool_id: tool name
            description:   description of this tool
        """

        Tools.__init__(self, tool_id, description)

        #: interface type (0 = demo, 4 = PCIe)
        self.interface_type = 0

        # create the controller Header object
        self.header = Header("Controller")
        self.header.set_header("controller", 2)

        #:  exposure time (secs)
        self.exposure_time = 0

        self.detpars = DetPars()

        #: True if the controller has been reset
        self.reset_flag = 0

        azcam.db.tools_init["controller"] = self
        azcam.db.tools_reset["controller"] = self

    def initialize(self):
        """
        Initialize tool.
        """

        self.is_initialized = 1

        return

    def reset(self):
        """
        Reset tool.
        """

        self.is_reset = 1

        return

    def set_roi(self):
        """
        XXXSets ROI parameters values in the controller based on focalplane parameters.
        """

        return

    def set_shutter_state(self, flag: bool = 0):
        """
        Sets the shutter state during an exposure.
        Args:
            flag: True open shutter during exposure, False close shutter during exposure
        """

        return

    def set_shutter(self, state: bool, shutter_id: int = 0):
        """
        Open or close controller shutter.
        Args:
            state: True to open shutter, False to close shutter
            shutter_id: shutter ID number
        """

        return

    def flush(self, cycles=1):
        """
        Flush or clear out the sensor.
        Returns after clearing is finished, which could take many seconds.
        Args:
            cycles: number of times to flush sensor
        """

        return

    def start_idle(self):
        """
        Start idle clocking.
        """

        return

    def stop_idle(self):
        """
        Stop idle clocking.
        """

        return

    def set_exposuretime(self, exposure_time: float):
        """
        Write the exposure time (in seconds) to the controller.
        """

        self.exposure_time = exposure_time

        return

    def get_exposuretime(self):
        """
        Return the exposure time from the controller (in seconds).
        """

        return self.exposure_time

    def get_pixels_remaining(self):
        """
        Return number of remaining pixels to be read (counts down).
        """

        return 0


class DetPars(object):
    """
    Defines detector parameters for controller shifting, image size, etc.
    These are updated by exposure.set_format, .set_focalplane, and .set_roi.
    Attributes:
        self.numpix_image (int):
        self.numcols_image (int):
        self.numrows_image (int):

        self.ns_total (int):
        self.ns_predark (int):
        self.ns_underscan (int):
        self.ns_overscan (int):
        self.np_total (int):
        self.np_predark (int):
        self.np_underscan (int):
        self.np_overscan (int):
        self.np_frametransfer (int):

        self.first_col (int):
        self.last_col (int):
        self.first_row (int):
        self.last_row (int):
        self.col_bin (int):
        self.row_bin (int):

        self.framet (int):
        self.xdata (int):
        self.xflush (int):
        self.xpreskip (int):
        self.xunderscan (int):
        self.xskip (int):
        self.xpostskip (int):
        self.xoverscan (int):
        self.ydata (int):
        self.yflush (int):
        self.ypreskip (int):
        self.yunderscan (int):
        self.yskip (int):
        self.ypostskip (int):
        self.yoverscan (int):

        self.coltotal (int):
        self.colusct (int):
        self.coluscw (int):
        self.coluscm (int):
        self.coloscw (int):
        self.coloscm (int):
        self.rowtotal (int):
        self.rowusct (int):
        self.rowuscw (int):
        self.rowuscm (int):
        self.rowoscw (int):
        self.rowoscm (int):
    """

    def __init__(self):
        """
        Create detpars database.
        """

        self.numpix_image = 0
        self.numcols_image = 0
        self.numrows_image = 0

        self.ns_total = 0
        self.ns_predark = 0
        self.ns_underscan = 0
        self.ns_overscan = 0
        self.np_total = 0
        self.np_predark = 0
        self.np_underscan = 0
        self.np_overscan = 0
        self.np_frametransfer = 0

        self.first_col = 0
        self.last_col = 0
        self.first_row = 0
        self.last_row = 0
        self.col_bin = 0
        self.row_bin = 0

        self.framet = 0
        self.xdata = 0
        self.xflush = 0
        self.xpreskip = 0
        self.xunderscan = 0
        self.xskip = 0
        self.xpostskip = 0
        self.xoverscan = 0
        self.ydata = 0
        self.yflush = 0
        self.ypreskip = 0
        self.yunderscan = 0
        self.yskip = 0
        self.ypostskip = 0
        self.yoverscan = 0

        # shifting parameters
        self.coltotal = 0
        self.colusct = 0
        self.coluscw = 0
        self.coluscm = 0
        self.coloscw = 0
        self.coloscm = 0
        self.rowtotal = 0
        self.rowusct = 0
        self.rowuscw = 0
        self.rowuscm = 0
        self.rowoscw = 0
        self.rowoscm = 0

        return



"""
*azcam.tools* contains the `Tools` base tool class.
"""

import azcam
import azcam.exceptions


class Tools(object):
    """
    Base class used by tools (controller, instrument, telescope, etc.).
    """

    def __init__(self, tool_id: str, description: str | None = None):
        """
        Args:
            tool_id: name used to reference the tool (controller, display, etc.)
            description: description of this tool
        """

        #: name used to reference the tool ("controller", "display", ...)
        self.tool_id: str = tool_id

        #: descriptive tool name
        self.description: str = ""
        if description is None:
            self.description = self.tool_id
        else:
            self.description = description

        #: 1 when tool is enabled
        self.is_enabled: int = 1

        #: 1 when tool has been initialized
        self.is_initialized: int = 0

        #: 1 when tool has been reset
        self.is_reset: int = 0

        #: verbosity for debug, >0 is more verbose
        self.verbosity = 0

        # save tool name
        azcam.db.tools.update({self.tool_id: self})

        # add tool to CLI
        azcam.db.cli.update({self.tool_id: self})



"""
*azcam.server* is imported to define server mode, create azcamserver
parameters dictionary, define a logger, and get local network information.
"""

import socket

import azcam
from azcam.logger import AzCamLogger
from azcam.parameters_server import ParametersServer
from azcam.database_server import AzcamDatabaseServer
from azcam.cmdserver import CommandServer
from azcam.api import API


def setup_server():
    azcam.db = AzcamDatabaseServer()  # overwrite default db

    # server mode
    azcam.db.set("servermode", "")

    # parameters
    azcam.db.parameters = ParametersServer()

    # logging
    azcam.db.logger = AzCamLogger()
    azcam.log = azcam.db.logger.log  # to allow azcam.log()

    # save this machine's hostname and ip address
    hostname = socket.gethostname()
    azcam.db.set("hostname", hostname)
    azcam.db.set("hostip", socket.gethostbyname(hostname))

    # tool_id's which are reset or initialized with exposure
    azcam.db.set("tools_reset", {})
    azcam.db.set("tools_init", {})

    # command server
    azcam.db.cmdserver = CommandServer()

    # define API tool
    azcam.db.api = API()
    azcam.db.cli["api"] = azcam.db.api

    return


# initial database but will ususally be overwritten by server or console
azcam.db: AzcamDatabaseServer = AzcamDatabaseServer()

# setup_server()



from dash import html, dcc, callback, Input, Output
import dash_bootstrap_components as dbc
import dash_daq as daq

import azcam


def detector_card():
    """
    Create detector card.
    """

    # First Column
    first_col_input = html.Div(
        [
            daq.NumericInput(id="first_col", label="First column", size=100, value=1),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("first_col", "value"),
        prevent_initial_call=True,
    )
    def first_col_callback(value):
        azcam.db.tools["exposure"].set_roi(first_col=int(value))
        return

    # Last Column
    last_col_input = html.Div(
        [
            daq.NumericInput(id="last_col", label="Last column", size=100, value=1),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("last_col", "value"),
        prevent_initial_call=True,
    )
    def last_col_callback(value):
        azcam.db.tools["exposure"].set_roi(last_col=int(value))
        return

    # Column binning
    col_bin_input = html.Div(
        [
            daq.NumericInput(id="col_bin", label="Column binning", size=100, value=1),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("col_bin", "value"),
        prevent_initial_call=True,
    )
    def col_bin_callback(value):
        azcam.db.tools["exposure"].set_roi(col_bin=int(value))
        return

    # First Row
    first_row_input = html.Div(
        [
            daq.NumericInput(id="first_row", label="First row", size=100, value=1),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("first_row", "value"),
        prevent_initial_call=True,
    )
    def first_row_callback(value):
        azcam.db.tools["exposure"].set_roi(first_row=int(value))
        return

    # Last Row
    last_row_input = html.Div(
        [
            daq.NumericInput(id="last_row", label="Last row", size=100, value=1),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("last_row", "value"),
        prevent_initial_call=True,
    )
    def last_row_callback(value):
        azcam.db.tools["exposure"].set_roi(last_row=int(value))
        return

    # Row binning
    row_bin_input = html.Div(
        [
            daq.NumericInput(id="row_bin", label="Row binning", size=100, value=1),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("row_bin", "value"),
        prevent_initial_call=True,
    )
    def row_bin_callback(value):
        azcam.db.tools["exposure"].set_roi(row_bin=int(value))
        return

    detector_button_group = dbc.ButtonGroup(
        [
            dbc.Button(
                "Set Full Frame", id="fullframe_btn", className="m-1", n_clicks=0
            ),
            dbc.Button("Apply All", id="detpars_btn", className="m-1", n_clicks=0),
        ],
        vertical=False,
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("fullframe_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_fullframe(n):
        azcam.db.tools["exposure"].roi_reset()

        return

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("detpars_btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def on_button_click_detpars(n):
        azcam.db.tools["exposure"].set_roi()

        return

    detector_card = dbc.Card(
        [
            dbc.CardHeader("Detector Parameters", style={"font-weight": "bold"}),
            dbc.CardBody(
                [
                    dbc.Row(
                        [
                            dbc.Col(first_col_input, width=2),
                            dbc.Col(last_col_input, width=2),
                            dbc.Col(col_bin_input, width=2),
                        ]
                    ),
                    dbc.Row(
                        [
                            dbc.Col(first_row_input, width=2),
                            dbc.Col(last_row_input, width=2),
                            dbc.Col(row_bin_input, width=2),
                        ]
                    ),
                    dbc.Row(
                        [
                            dbc.Col(detector_button_group),
                        ]
                    ),
                ]
            ),
        ]
    )

    return detector_card



"""
Contains the display tool.
This tool is often implemented both by server and console.
"""

import azcam
from azcam.header import Header, ObjectHeaderMethods
from azcam.tools.tools import Tools
import azcam.exceptions


class Display(Tools, ObjectHeaderMethods):
    """
    The base display tool for server and consoles.
    Usually implemented as the "display" tool.
    """

    def __init__(self, tool_id="display", description=None):
        Tools.__init__(self, tool_id, description)

        # create the display Header object
        self.header = Header("Display")
        self.header.set_header("display", 2)

        # set default display server
        self.default_display = 0

        # allow initialization of display on server in exposure
        try:
            azcam.db.tools_init["display"] = self
        except AttributeError:
            pass

    def initialize(self):
        """
        Initialize display.
        """

        if self.is_initialized:
            return

        if not self.is_enabled:
            azcam.exceptions.warning("Display is not enabled")
            return

        self.set_display(self.default_display)

        return

    def reset(self):
        """
        Reset tool.
        """

        self.is_reset = 1

        return

    def set_display(self, display_number):
        """
        Set the current display by number.

        :param int display_number: Number for display to be used (0->N)
        :return None:
        """

        return

    def display(self, image, extension_number=-1):
        """
        Display a file on the image display.
        If specified for an MEF file, only extension_number is displayed.

        :param image: a filename or an image object
        :param int extension_number: FITS extension number of image, -1 for all
        :return None:
        """

        raise NotImplementedError



"""
*azcam.sockets* contains azcam's `SocketInterface` class used by many interfaces.
"""

import shlex
import socket
import time
import threading

import azcam
import azcam.exceptions


class SocketInterface(object):
    """
    The azcam socket server interface.
    """

    def __init__(self, host: str = "", port: int = 0) -> None:
        """
        Create socket connection instance.

        Args:
            host: host name of remote computer.
            port: port number of server on remote computer.
        """

        #: host name or IP address of remote server
        self.host: str = host
        #: port number of remote server
        self.port: int = port

        if host != "":
            self.host = host
        if port != 0:
            self.port = port

        #: true when connection is active
        self.connected: int = 0
        #: termination string for commands
        self.terminator = "\r\n"
        #: last response from server
        self.last_response = ""
        #: last command send to server
        self.last_command = ""
        #: true to try reopening socket a reset socket when trying to connect
        self.reopen = 0
        #: connection timeout in seconds
        self.timeout = 0

        #: socket instance
        self.socket: object = None

        #: thread lock
        self.lock = threading.Lock()

    def open(self) -> bool:
        """
        Open a socket connection to the server.
        Creates the socket and makes a connection.
        Returns:
            True if socket is already open or if it is opened here.
        """

        # check if socket is already open
        if self.socket is not None:
            return True

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 0)
        try:
            self.socket.settimeout(0.1)
            self.socket.connect((self.host, int(self.port)))
            self.connected = True
        except Exception:
            self.close()
            self.connected = False
            return False

        # check if there is a welcome message
        self.set_timeout(0.2)
        try:
            welcome = self.recv()
        except socket.timeout:
            welcome = ""
        self.set_timeout(0)  # initally no timeout

        if len(welcome) > 0:
            azcam.log(welcome)

        return True

    def close(self) -> None:
        """
        Close the socket connection to the server.
        """

        if self.socket is not None:
            try:
                self.socket.close()
            except Exception:
                pass  # never error on close
        self.socket = None
        self.connected = False

        return

    def command(self, command: str, terminator: str = "\n") -> list:
        """
        Sends a command to the server and receives a reply back.

        Args:
            command: command string to send
            terminator: termination string to append the command
        Returns:
            tokenized list of the server's reply.
        """

        with self.lock:
            if not self.open():
                raise azcam.exceptions.AzcamError(
                    "could not open connection to server", error_code=2
                )

            self.send(command, terminator)
            reply = self.recv(-1, "\n")

        self.last_response = reply

        if command not in ["exposure.get_status"]:
            reply = shlex.split(reply)
        else:
            reply = [reply.strip()]

        return reply

    def send(self, command: str, terminator: str = "\n") -> None:
        """
        Send a command string to the server.
        Args:
            command: command string to send
            terminator: termination string to append the command
        """

        self.last_command = command
        try:
            self.socket.send(str.encode(command + terminator))
        except ConnectionResetError:
            if self.reopen:
                self.close()
                self.open()
                self.socket.send(str.encode(command + terminator))
            else:
                raise

        return

    def recv(self, length: int = -1, terminator: str = "\n") -> str:
        """
        Receives a reply from a server.
        Terminates socket read when length bytes are received or when the terminator is received.
        Args:
            length: number of bytes to receive. -1 means receive through terminator.
            terminator: terminator string.
        """

        # receive Length bytes
        if length != -1:
            reply = self.socket.recv(length).decode()
            return reply

        # read socket until Terminator found
        msg = ""
        loop = 0
        while True:
            try:
                msg += self.socket.recv(1024).decode()
            except ConnectionAbortedError:
                raise azcam.exceptions.AzcamError("Connection aborted")
            if len(msg) == 0:
                pass
            elif msg[-1] == terminator:  # found terminator at end
                break
            # check for infinite loop
            loop += 1
            if loop > 1024:
                raise azcam.exceptions.AzcamError("Socket recv loop max retry exceeded")
            time.sleep(0.005)

        if len(msg) < 2:
            raise azcam.exceptions.AzcamError("Invalid command response received")
        if msg[-2] == "r":
            reply = msg[:-2]  # remove CR/LF
        else:
            reply = msg[:-1]  # remove LF

        return reply

    def set_timeout(self, timeout: float = -1.0) -> None:
        """
        Set the receive timeout for socket connection.
        Args:
            timeout: timeout value in seconds. <0 means use current value, =0 means no timeout.
        """

        if timeout < 0:
            pass
        elif timeout == 0:
            self.timeout = None
        else:
            self.timeout = float(timeout)

        self.socket.settimeout(self.timeout)

        return

    def test(self) -> str:
        """
        Sends and receives a test message over the socket connection.
        """

        message = "this is a test message string"

        return self.command(f'Echo "{message}"')



from dash import html, dcc, callback, Input, Output
import dash_bootstrap_components as dbc
import dash_daq as daq

import azcam


def options_card():
    """
    Create options card.
    """

    # flush sensor
    flush_input = html.Div(
        [
            dbc.Checkbox(
                id="flush_sensor",
                label="Clear sensor before each exposure",
                value=True,
            ),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("flush_sensor", "value"),
        prevent_initial_call=True,
    )
    def flush_sensor_callback(value):
        azcam.db.tools["exposure"].flush_array = int(value)
        return

    # display image
    display_image_input = html.Div(
        [
            dbc.Checkbox(
                id="display_image",
                label="Display image from server",
                value=True,
            ),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("display_image", "value"),
        prevent_initial_call=True,
    )
    def display_image_callback(value):
        azcam.db.tools["exposure"].display_image = int(value)
        return

    # enable instrument
    enable_instrument_input = html.Div(
        [
            dbc.Checkbox(
                id="enable_instrument",
                label="Enable instrument",
                value=True,
            ),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("enable_instrument", "value"),
        prevent_initial_call=True,
    )
    def enable_instrument_callback(value):
        try:
            azcam.db.tools["instrument"].is_enabled = int(value)
        except Exception:
            azcam.db.expserver.set_message("instrument not defined")
        return

    # enable telescope
    enable_telescope_input = html.Div(
        [
            dbc.Checkbox(
                id="enable_telescope",
                label="Enable telescope",
                value=True,
            ),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("enable_telescope", "value"),
        prevent_initial_call=True,
    )
    def enable_telescope_callback(value):
        try:
            azcam.db.tools["telescope"].is_enabled = int(value)
        except Exception:
            azcam.db.expserver.set_message("telescope is not defined")
        return

    # auto title images
    auto_title_input = html.Div(
        [
            dbc.Checkbox(
                id="auto_title",
                label="Auto title images",
                value=True,
            ),
        ]
    )

    @callback(
        Output("hidden_div", "children", allow_duplicate=True),
        Input("auto_title", "value"),
        prevent_initial_call=True,
    )
    def auto_title_callback(value):
        azcam.db.tools["exposure"].auto_title = int(value)
        return

    options_card = dbc.Card(
        [
            dbc.CardHeader("Options", style={"font-weight": "bold"}),
            dbc.CardBody(
                [
                    flush_input,
                    display_image_input,
                    enable_instrument_input,
                    enable_telescope_input,
                    auto_title_input,
                ]
            ),
        ]
    )

    return options_card



"""
Contains the ControllerASCOM class.
"""

import time
from datetime import datetime

try:
    import win32com.client
except Exception as e:
    print(e)

import azcam
import azcam.exceptions
from azcam.tools.controller import Controller


class ControllerASCOM(Controller):
    """
    The controller class for cameras using ASCOM.
    """

    def __init__(self, tool_id="controller", description=None):
        super().__init__(tool_id, description)

        self.controller_class = "ASCOM"

        # video gain flag
        self.video_gain = 1

        # True if image data from ControllserServer is binary (not a file)
        self.binary_image_data = 1

        self.shutter_state = 0

        # ASCOM
        self.driver = ""
        self.camera = None

    def initialize(self):
        """
        Initialize the controller interface.
        """

        if self.is_initialized:
            return

        azcam.log("initializing camera")

        """
        import win32com
        driver = "ASCOM.ASICamera2.Camera"
        camera = win32com.client.Dispatch(driver)
        camera.connected = True
        camera.NumX = self.nx
        camera.NumY = self.ny
        camera.BinX = 1
        camera.BinY = 1
        camera.StartExposure(1, 1)
        time.sleep(3)
        flag = camera.ImageReady
        print("ready", flag)
        x = win32com.client.Dispatch("ASCOM.Utilities.Chooser")
        x.DeviceType = "Camera"
        driver = x.Choose(None)
        """

        self.camera = win32com.client.Dispatch(self.driver)
        self.camera.connected = True

        """
        self.camera.NumX = self.nx
        self.camera.NumY = self.ny
        self.camera.BinX = 1
        self.camera.BinY = 1
        """

        self.is_initialized = 1

        return

    def reset(self):
        """
        Reset the controller.
        """

        # reset flag even is system has previously been reset
        self.is_reset = 0

        self.header.delete_all_keywords()

        self.is_reset = 1

        self.set_roi()

        self.set_exposuretime(azcam.db.tools["exposure"].exposure_time)

        return

    def set_shutter_state(self, state):
        """
        Open or close controller shutter.
        """

        if state:
            self.shutter_state = 1
        else:
            self.shutter_state = 0

        return

    def set_exposuretime(self, exposure_time):
        """
        Write the exposure time (in seconds) to the controller.
        """

        exposure_time = float(exposure_time)
        et = max(0.000001, exposure_time)

        self.exposure_time = et

        return

    def get_exposuretime(self):
        return self.exposure_time

    def start_exposure(self):
        """
        Start exposure (integration).
        """

        self.start_time = datetime.now()

        try:
            self.camera.StartExposure(self.exposure_time, self.shutter_state)
        except Exception as e:
            print(f"Camera error starting exposure: {e}")

        return

    def is_imageready(self, wait=0):
        """
        Return True if image is ready.
        """

        flag = self.camera.ImageReady

        count = 0
        if wait:
            while not flag:
                if count > 200:
                    raise azcam.exceptions.AzcamError("Camera timeout reading image")
                time.sleep(0.05)
                count += 1
                # print("waiting...")
                flag = self.camera.ImageReady
            # print("ready")

        return flag

    def update_exposuretime_remaining(self):
        """
        Return remaining exposure time (in seconds).
        """

        et = datetime.now() - self.start_time
        et = self.exposure_time - et.seconds
        if et < 0:
            et = 0

        return et

    def get_pixels_remaining(self):
        """
        Return number of remaining pixels to be read (counts down).
        """

        return 0

    def set_roi(self):
        self.camera.BinX = min(self.detpars.col_bin, self.camera.MaxBinX)
        self.camera.BinY = min(self.detpars.row_bin, self.camera.MaxBinY)

        # binned pixels
        self.camera.NumX = min(self.detpars.numcols_image, self.camera.CameraXSize)
        self.camera.NumY = min(self.detpars.numrows_image, self.camera.CameraYSize)

        # zero based
        self.camera.StartX = self.detpars.first_col / self.detpars.col_bin - 1
        self.camera.StartY = self.detpars.first_row / self.detpars.row_bin - 1

        return
